#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x8fe3664f

# Compiled with Coconut version 1.3.1-post_dev17 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

# Copyright Arne Bachmann
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import builtins  # line 4
import enum  # line 4
import json  # line 4
import logging  # line 4
import os  # line 4
import shutil  # line 4
sys = _coconut_sys  # line 4
import time  # line 4
import traceback  # line 4
import unittest  # line 4
import uuid  # line 4
from io import BytesIO  # line 5
from io import BufferedRandom  # line 5
from io import TextIOWrapper  # line 5
try:  # Python 3  # line 6
    from unittest import mock  # Python 3  # line 6
except:  # installed via pip  # line 7
    import mock  # installed via pip  # line 7
try:  # only required for mypy  # line 8
    from typing import Any  # only required for mypy  # line 8
    from typing import List  # only required for mypy  # line 8
    from typing import Union  # only required for mypy  # line 8
except:  # line 9
    pass  # line 9

testFolder = os.path.abspath(os.path.join(os.getcwd(), "test"))  # line 11
import configr  # optional dependency  # line 12
os.environ["TEST"] = testFolder  # needed to mock configr library calls in sos  # line 13
import sos  # import of package, not file  # line 14
sos.defaults["defaultbranch"] = "trunk"  # because sos.main() is never called  # line 15

def sync():  # line 17
    if sys.version_info[:2] >= (3, 3):  # line 18
        os.sync()  # line 18


def determineFilesystemTimeResolution() -> 'float':  # line 21
    name = str(uuid.uuid4())  # type: str  # line 22
    with open(name, "w") as fd:  # create temporary file  # line 23
        fd.write("x")  # create temporary file  # line 23
    mt = os.stat(name).st_mtime  # type: float  # get current timestamp  # line 24
    while os.stat(name).st_mtime == mt:  # wait until timestamp modified  # line 25
        time.sleep(0.05)  # to avoid 0.00s bugs (came up some time for unknown reasons)  # line 26
        with open(name, "w") as fd:  # line 27
            fd.write("x")  # line 27
    mt, start, _count = os.stat(name).st_mtime, time.time(), 0  # line 28
    while os.stat(name).st_mtime == mt:  # now cound and measure time until modified again  # line 29
        time.sleep(0.05)  # line 30
        _count += 1  # line 31
        with open(name, "w") as fd:  # line 32
            fd.write("x")  # line 32
    os.unlink(name)  # line 33
    fsprecision = round(time.time() - start, 2)  # type: float  # line 34
    print("File system timestamp precision is %.2fs; wrote to the file %d times during that time" % (fsprecision, _count))  # line 35
    return fsprecision  # line 36


FS_PRECISION = determineFilesystemTimeResolution() * 1.05  # line 39


@_coconut_tco  # line 42
def debugTestRunner(post_mortem=None):  # line 42
    ''' Unittest runner doing post mortem debugging on failing tests. '''  # line 43
    import pdb  # line 44
    if post_mortem is None:  # line 45
        post_mortem = pdb.post_mortem  # line 45
    class DebugTestResult(unittest.TextTestResult):  # line 46
        def addError(_, test, err):  # called before tearDown()  # line 47
            traceback.print_exception(*err)  # line 48
            post_mortem(err[2])  # line 49
            super(DebugTestResult, _).addError(test, err)  # line 50
        def addFailure(_, test, err):  # line 51
            traceback.print_exception(*err)  # line 52
            post_mortem(err[2])  # line 53
            super(DebugTestResult, _).addFailure(test, err)  # line 54
    return _coconut_tail_call(unittest.TextTestRunner, resultclass=DebugTestResult)  # line 55

@_coconut_tco  # line 57
def wrapChannels(func: '_coconut.typing.Callable[..., Any]'):  # line 57
    ''' Wrap function call to capture and return strings emitted on stdout and stderr. '''  # line 58
    oldv = sys.argv  # line 59
    buf = TextIOWrapper(BufferedRandom(BytesIO(b"")), encoding="utf-8")  # line 60
    sys.stdout = sys.stderr = buf  # line 61
    handler = logging.StreamHandler(buf)  # line 62
    logging.getLogger().addHandler(handler)  # line 63
    try:  # capture output into buf  # line 64
        func()  # capture output into buf  # line 64
    except Exception as E:  # line 65
        buf.write(str(E) + "\n")  # line 65
        traceback.print_exc(file=buf)  # line 65
    except SystemExit as F:  # line 66
        buf.write("EXIT CODE %s" % F.code + "\n")  # line 66
        traceback.print_exc(file=buf)  # line 66
    logging.getLogger().removeHandler(handler)  # line 67
    sys.argv, sys.stdout, sys.stderr = oldv, sys.__stdout__, sys.__stderr__  # TODO when run using pythonw.exe and/or no console, these could be None  # line 68
    buf.seek(0)  # line 69
    return _coconut_tail_call(buf.read)  # line 70

def mockInput(datas: '_coconut.typing.Sequence[str]', func) -> 'Any':  # line 72
    with mock.patch("builtins.input", side_effect=datas):  # line 73
        return func()  # line 73

def setRepoFlag(name: 'str', value: 'bool'):  # line 75
    with open(sos.metaFolder + os.sep + sos.metaFile, "r") as fd:  # line 76
        flags, branches, config = json.loads(fd.read())  # line 76
    flags[name] = value  # line 77
    with open(sos.metaFolder + os.sep + sos.metaFile, "w") as fd:  # line 78
        fd.write(json.dumps((flags, branches, config)))  # line 78

def checkRepoFlag(name: 'str', flag: '_coconut.typing.Optional[bool]'=None, value: '_coconut.typing.Optional[Any]'=None) -> 'bool':  # line 80
    with open(sos.metaFolder + os.sep + sos.metaFile, "r") as fd:  # line 81
        flags, branches, config = json.loads(fd.read())  # line 81
    return (name in flags and flags[name] == flag) if flag is not None else (name in config and config[name] == value)  # line 82


class Tests(unittest.TestCase):  # line 85
    ''' Entire test suite. '''  # line 86

    def setUp(_):  # line 88
        for entry in os.listdir(testFolder):  # cannot remove testFolder on Windows when using TortoiseSVN as VCS  # line 89
            resource = os.path.join(testFolder, entry)  # line 90
            shutil.rmtree(resource) if os.path.isdir(resource) else os.unlink(resource)  # line 91
        os.chdir(testFolder)  # line 92


    def assertAllIn(_, what: '_coconut.typing.Sequence[str]', where: 'Union[str, List[str]]'):  # line 95
        for w in what:  # line 96
            _.assertIn(w, where)  # line 96

    def assertAllNotIn(_, what: '_coconut.typing.Sequence[str]', where: 'Union[str, List[str]]'):  # line 98
        for w in what:  # line 99
            _.assertNotIn(w, where)  # line 99

    def assertInAll(_, what: 'str', where: '_coconut.typing.Sequence[str]'):  # line 101
        for w in where:  # line 102
            _.assertIn(what, w)  # line 102

    def assertInAny(_, what: 'str', where: '_coconut.typing.Sequence[str]'):  # line 104
        _.assertTrue(any((what in w for w in where)))  # line 104

    def assertNotInAny(_, what: 'str', where: '_coconut.typing.Sequence[str]'):  # line 106
        _.assertFalse(any((what in w for w in where)))  # line 106

    def createFile(_, number: 'Union[int, str]', contents: 'str'="x" * 10, prefix: '_coconut.typing.Optional[str]'=None):  # line 108
        if prefix and not os.path.exists(prefix):  # line 109
            os.makedirs(prefix)  # line 109
        with open(("." if prefix is None else prefix) + os.sep + (("file%d" % number) if isinstance(number, int) else number), "wb") as fd:  # line 110
            fd.write(contents if isinstance(contents, bytes) else contents.encode("cp1252"))  # line 110

    def existsFile(_, number: 'Union[int, str]', expectedContents: 'bytes'=None) -> 'bool':  # line 112
        if not os.path.exists(("." + os.sep + "file%d" % number) if isinstance(number, int) else number):  # line 113
            return False  # line 113
        if expectedContents is None:  # line 114
            return True  # line 114
        with open(("." + os.sep + "file%d" % number) if isinstance(number, int) else number, "rb") as fd:  # line 115
            return fd.read() == expectedContents  # line 115

    def testAccessor(_):  # line 117
        a = sos.Accessor({"a": 1})  # line 118
        _.assertEqual((1, 1), (a["a"], a.a))  # line 119

    def testGetAnyOfmap(_):  # line 121
        _.assertEqual(2, sos.getAnyOfMap({"a": 1, "b": 2}, ["x", "b"]))  # line 122
        _.assertIsNone(sos.getAnyOfMap({"a": 1, "b": 2}, []))  # line 123

    def testAjoin(_):  # line 125
        _.assertEqual("a1a2", sos.ajoin("a", ["1", "2"]))  # line 126
        _.assertEqual("* a\n* b", sos.ajoin("* ", ["a", "b"], "\n"))  # line 127

    def testFindChanges(_):  # line 129
        m = sos.Metadata(os.getcwd())  # line 130
        try:  # line 131
            sos.config(["set", "texttype", "*"])  # line 131
        except SystemExit as E:  # line 132
            _.assertEqual(0, E.code)  # line 132
        try:  # will be stripped from leading paths anyway  # line 133
            sos.config(["set", "ignores", "test/*.cfg;D:\\apps\\*.cfg.bak"])  # will be stripped from leading paths anyway  # line 133
        except SystemExit as E:  # line 134
            _.assertEqual(0, E.code)  # line 134
        m = sos.Metadata(os.getcwd())  # reload from file system  # line 135
        for file in [f for f in os.listdir() if f.endswith(".bak")]:  # remove configuration file  # line 136
            os.unlink(file)  # remove configuration file  # line 136
        _.createFile(1, "1")  # line 137
        m.createBranch(0)  # line 138
        _.assertEqual(1, len(m.paths))  # line 139
        time.sleep(FS_PRECISION)  # time required by filesystem time resolution issues  # line 140
        _.createFile(1, "2")  # modify existing file  # line 141
        _.createFile(2, "2")  # add another file  # line 142
        m.loadCommit(0, 0)  # line 143
        changes = m.findChanges()  # detect time skew  # line 144
        _.assertEqual(1, len(changes.additions))  # line 145
        _.assertEqual(0, len(changes.deletions))  # line 146
        _.assertEqual(1, len(changes.modifications))  # line 147
        m.paths.update(changes.additions)  # line 148
        m.paths.update(changes.modifications)  # line 149
        _.createFile(2, "12")  # modify file again  # line 150
        changes = m.findChanges(0, 1)  # by size, creating new commit  # line 151
        _.assertEqual(0, len(changes.additions))  # line 152
        _.assertEqual(0, len(changes.deletions))  # line 153
        _.assertEqual(1, len(changes.modifications))  # line 154
        _.assertTrue(os.path.exists(sos.branchFolder(0, 1)))  # line 155
        _.assertTrue(os.path.exists(sos.branchFolder(0, 1, file="03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2")))  # line 156

    def testPatternPaths(_):  # line 158
        sos.offline(options=["--track"])  # line 159
        os.mkdir("sub")  # line 160
        _.createFile("sub" + os.sep + "file1", "sdfsdf")  # line 161
        sos.add("sub", "sub/file?")  # line 162
        sos.commit("test")  # should pick up sub/file1 pattern  # line 163
        _.assertEqual(2, len(os.listdir(os.path.join(sos.metaFolder, "b0", "r1"))))  # sub/file1 was added  # line 164
        _.createFile(1)  # line 165
        try:  # should not commit anything, as the file in base folder doesn't match the tracked pattern  # line 166
            sos.commit("nothing")  # should not commit anything, as the file in base folder doesn't match the tracked pattern  # line 166
            _.fail()  # should not commit anything, as the file in base folder doesn't match the tracked pattern  # line 166
        except:  # line 167
            pass  # line 167

    def testTokenizeGlobPattern(_):  # line 169
        _.assertEqual([], sos.tokenizeGlobPattern(""))  # line 170
        _.assertEqual([sos.GlobBlock(False, "*", 0)], sos.tokenizeGlobPattern("*"))  # line 171
        _.assertEqual([sos.GlobBlock(False, "*", 0), sos.GlobBlock(False, "???", 1)], sos.tokenizeGlobPattern("*???"))  # line 172
        _.assertEqual([sos.GlobBlock(True, "x", 0), sos.GlobBlock(False, "*", 1), sos.GlobBlock(True, "x", 2)], sos.tokenizeGlobPattern("x*x"))  # line 173
        _.assertEqual([sos.GlobBlock(True, "x", 0), sos.GlobBlock(False, "*", 1), sos.GlobBlock(False, "??", 2), sos.GlobBlock(False, "*", 4), sos.GlobBlock(True, "x", 5)], sos.tokenizeGlobPattern("x*??*x"))  # line 174
        _.assertEqual([sos.GlobBlock(False, "?", 0), sos.GlobBlock(True, "abc", 1), sos.GlobBlock(False, "*", 4)], sos.tokenizeGlobPattern("?abc*"))  # line 175

    def testTokenizeGlobPatterns(_):  # line 177
        try:  # because number of literal strings differs  # line 178
            sos.tokenizeGlobPatterns("x*x", "x*")  # because number of literal strings differs  # line 178
            _.fail()  # because number of literal strings differs  # line 178
        except:  # line 179
            pass  # line 179
        try:  # because glob patterns differ  # line 180
            sos.tokenizeGlobPatterns("x*", "x?")  # because glob patterns differ  # line 180
            _.fail()  # because glob patterns differ  # line 180
        except:  # line 181
            pass  # line 181
        try:  # glob patterns differ, regardless of position  # line 182
            sos.tokenizeGlobPatterns("x*", "?x")  # glob patterns differ, regardless of position  # line 182
            _.fail()  # glob patterns differ, regardless of position  # line 182
        except:  # line 183
            pass  # line 183
        sos.tokenizeGlobPatterns("x*", "*x")  # succeeds, because glob patterns match (differ only in position)  # line 184
        sos.tokenizeGlobPatterns("*xb?c", "*x?bc")  # succeeds, because glob patterns match (differ only in position)  # line 185
        try:  # succeeds, because glob patterns match (differ only in position)  # line 186
            sos.tokenizeGlobPatterns("a???b*", "ab???*")  # succeeds, because glob patterns match (differ only in position)  # line 186
            _.fail()  # succeeds, because glob patterns match (differ only in position)  # line 186
        except:  # line 187
            pass  # line 187

    def testConvertGlobFiles(_):  # line 189
        _.assertEqual(["xxayb", "aacb"], [r[1] for r in sos.convertGlobFiles(["axxby", "aabc"], *sos.tokenizeGlobPatterns("a*b?", "*a?b"))])  # line 190
        _.assertEqual(["1qq2ww3", "1abcbx2xbabc3"], [r[1] for r in sos.convertGlobFiles(["qqxbww", "abcbxxbxbabc"], *sos.tokenizeGlobPatterns("*xb*", "1*2*3"))])  # line 191

    def testFolderRemove(_):  # line 193
        m = sos.Metadata(os.getcwd())  # line 194
        _.createFile(1)  # line 195
        _.createFile("a", prefix="sub")  # line 196
        sos.offline()  # line 197
        _.createFile(2)  # line 198
        os.unlink("sub" + os.sep + "a")  # line 199
        os.rmdir("sub")  # line 200
        changes = sos.changes()  # TODO replace by output check  # line 201
        _.assertEqual(1, len(changes.additions))  # line 202
        _.assertEqual(0, len(changes.modifications))  # line 203
        _.assertEqual(1, len(changes.deletions))  # line 204
        _.createFile("a", prefix="sub")  # line 205
        changes = sos.changes()  # line 206
        _.assertEqual(0, len(changes.deletions))  # line 207

    def testSwitchConflict(_):  # line 209
        sos.offline(options=["--strict"])  # (r0)  # line 210
        _.createFile(1)  # line 211
        sos.commit()  # add file (r1)  # line 212
        os.unlink("file1")  # line 213
        sos.commit()  # remove (r2)  # line 214
        _.createFile(1, "something else")  # line 215
        sos.commit()  # (r3)  # line 216
        sos.switch("/1")  # updates file1 - marked as MOD, because mtime was changed  # line 217
        _.existsFile(1, "x" * 10)  # line 218
        sos.switch("/2", ["--force"])  # remove file1 requires --force, because size/content (or mtime in non-strict mode) is different to head of branch  # line 219
        sos.switch("/0")  # do nothing, as file1 is already removed  # line 220
        sos.switch("/1")  # add file1 back  # line 221
        sos.switch("/", ["--force"])  # requires force because changed vs. head of branch  # line 222
        _.existsFile(1, "something else")  # line 223

    def testComputeSequentialPathSet(_):  # line 225
        os.makedirs(sos.branchFolder(0, 0))  # line 226
        os.makedirs(sos.branchFolder(0, 1))  # line 227
        os.makedirs(sos.branchFolder(0, 2))  # line 228
        os.makedirs(sos.branchFolder(0, 3))  # line 229
        os.makedirs(sos.branchFolder(0, 4))  # line 230
        m = sos.Metadata(os.getcwd())  # line 231
        m.branch = 0  # line 232
        m.commit = 2  # line 233
        m.saveBranches()  # line 234
        m.paths = {"./a": sos.PathInfo("", 0, 0, "")}  # line 235
        m.saveCommit(0, 0)  # initial  # line 236
        m.paths["./a"] = sos.PathInfo("", 1, 0, "")  # line 237
        m.saveCommit(0, 1)  # mod  # line 238
        m.paths["./b"] = sos.PathInfo("", 0, 0, "")  # line 239
        m.saveCommit(0, 2)  # add  # line 240
        m.paths["./a"] = sos.PathInfo("", None, 0, "")  # line 241
        m.saveCommit(0, 3)  # del  # line 242
        m.paths["./a"] = sos.PathInfo("", 2, 0, "")  # line 243
        m.saveCommit(0, 4)  # readd  # line 244
        m.commits = {i: sos.CommitInfo(i, 0, None) for i in range(5)}  # line 245
        m.saveBranch(0)  # line 246
        m.computeSequentialPathSet(0, 4)  # line 247
        _.assertEqual(2, len(m.paths))  # line 248

    def testParseRevisionString(_):  # line 250
        m = sos.Metadata(os.getcwd())  # line 251
        m.branch = 1  # line 252
        m.commits = {0: 0, 1: 1, 2: 2}  # line 253
        _.assertEqual((1, 3), m.parseRevisionString("3"))  # line 254
        _.assertEqual((2, 3), m.parseRevisionString("2/3"))  # line 255
        _.assertEqual((1, -1), m.parseRevisionString(None))  # line 256
        _.assertEqual((1, -1), m.parseRevisionString(""))  # line 257
        _.assertEqual((2, -1), m.parseRevisionString("2/"))  # line 258
        _.assertEqual((1, -2), m.parseRevisionString("/-2"))  # line 259
        _.assertEqual((1, -1), m.parseRevisionString("/"))  # line 260

    def testOfflineEmpty(_):  # line 262
        os.mkdir("." + os.sep + sos.metaFolder)  # line 263
        try:  # line 264
            sos.offline("trunk")  # line 264
            _.fail()  # line 264
        except SystemExit as E:  # line 265
            _.assertEqual(1, E.code)  # line 265
        os.rmdir("." + os.sep + sos.metaFolder)  # line 266
        sos.offline("test")  # line 267
        _.assertIn(sos.metaFolder, os.listdir("."))  # line 268
        _.assertAllIn(["b0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder))  # line 269
        _.assertAllIn(["r0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 270
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder)))  # only branch folder and meta data file  # line 271
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0")))  # only commit folder and meta data file  # line 272
        _.assertEqual(1, len(os.listdir(sos.branchFolder(0, 0))))  # only meta data file  # line 273

    def testOfflineWithFiles(_):  # line 275
        _.createFile(1, "x" * 100)  # line 276
        _.createFile(2)  # line 277
        sos.offline("test")  # line 278
        _.assertAllIn(["file1", "file2", sos.metaFolder], os.listdir("."))  # line 279
        _.assertAllIn(["b0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder))  # line 280
        _.assertAllIn(["r0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 281
        _.assertAllIn([sos.metaFile, "03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2", "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa"], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0" + os.sep + "r0"))  # line 282
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder)))  # only branch folder and meta data file  # line 283
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0")))  # only commit folder and meta data file  # line 284
        _.assertEqual(3, len(os.listdir(sos.branchFolder(0, 0))))  # only meta data file plus branch base file copies  # line 285

    def testBranch(_):  # line 287
        _.createFile(1, "x" * 100)  # line 288
        _.createFile(2)  # line 289
        sos.offline("test")  # b0/r0  # line 290
        sos.branch("other")  # b1/r0  # line 291
        _.assertAllIn(["b0", "b1", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder))  # line 292
        _.assertEqual(list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))), list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b1"))))  # line 293
        _.assertEqual(list(sorted(os.listdir(sos.branchFolder(0, 0)))), list(sorted(os.listdir(sos.branchFolder(1, 0)))))  # line 295
        _.createFile(1, "z")  # modify file  # line 297
        sos.branch()  # b2/r0  branch to unnamed branch with modified file tree contents  # line 298
        _.assertNotEqual(os.stat("." + os.sep + sos.metaFolder + os.sep + "b1" + os.sep + "r0" + os.sep + "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa").st_size, os.stat("." + os.sep + sos.metaFolder + os.sep + "b2" + os.sep + "r0" + os.sep + "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa").st_size)  # line 299
        _.createFile(3, "z")  # line 301
        sos.branch("from_last_revision", ["--last", "--stay"])  # b3/r0 create copy of other file1,file2 and don't switch  # line 302
        _.assertEqual(list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b3" + os.sep + "r0"))), list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b2" + os.sep + "r0"))))  # line 303
# Check sos.status output which branch is marked


    def testComittingAndChanges(_):  # line 308
        _.createFile(1, "x" * 100)  # line 309
        _.createFile(2)  # line 310
        sos.offline("test")  # line 311
        changes = sos.changes()  # line 312
        _.assertEqual(0, len(changes.additions))  # line 313
        _.assertEqual(0, len(changes.deletions))  # line 314
        _.assertEqual(0, len(changes.modifications))  # line 315
        _.createFile(1, "z")  # size change  # line 316
        changes = sos.changes()  # line 317
        _.assertEqual(0, len(changes.additions))  # line 318
        _.assertEqual(0, len(changes.deletions))  # line 319
        _.assertEqual(1, len(changes.modifications))  # line 320
        sos.commit("message")  # line 321
        _.assertAllIn(["r0", "r1", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 322
        _.assertAllIn([sos.metaFile, "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa"], os.listdir(sos.branchFolder(0, 1)))  # line 323
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # no further files, only the modified one  # line 324
        _.assertEqual(1, len(sos.changes("/0").modifications))  # vs. explicit revision on current branch  # line 325
        _.assertEqual(1, len(sos.changes("0/0").modifications))  # vs. explicit branch/revision  # line 326
        _.createFile(1, "")  # modify to empty file, mentioned in meta data, but not stored as own file  # line 327
        os.unlink("file2")  # line 328
        changes = sos.changes()  # line 329
        _.assertEqual(0, len(changes.additions))  # line 330
        _.assertEqual(1, len(changes.deletions))  # line 331
        _.assertEqual(1, len(changes.modifications))  # line 332
        sos.commit("modified")  # line 333
        _.assertEqual(1, len(os.listdir(sos.branchFolder(0, 2))))  # no additional files, only mentions in metadata  # line 334
        try:  # expecting Exit due to no changes  # line 335
            sos.commit("nothing")  # expecting Exit due to no changes  # line 335
            _.fail()  # expecting Exit due to no changes  # line 335
        except:  # line 336
            pass  # line 336

    def testGetBranch(_):  # line 338
        m = sos.Metadata(os.getcwd())  # line 339
        m.branch = 1  # current branch  # line 340
        m.branches = {0: sos.BranchInfo(0, 0, "trunk")}  # line 341
        _.assertEqual(27, m.getBranchByName(27))  # line 342
        _.assertEqual(0, m.getBranchByName("trunk"))  # line 343
        _.assertEqual(1, m.getBranchByName(""))  # split from "/"  # line 344
        _.assertIsNone(m.getBranchByName("unknown"))  # line 345
        m.commits = {0: sos.CommitInfo(0, 0, "bla")}  # line 346
        _.assertEqual(13, m.getRevisionByName("13"))  # line 347
        _.assertEqual(0, m.getRevisionByName("bla"))  # line 348
        _.assertEqual(-1, m.getRevisionByName(""))  # split from "/"  # line 349

    def testTagging(_):  # line 351
        m = sos.Metadata(os.getcwd())  # line 352
        sos.offline()  # line 353
        _.createFile(111)  # line 354
        sos.commit("tag", ["--tag"])  # line 355
        out = wrapChannels(lambda: sos.log()).replace("\r", "").split("\n")  # line 356
        _.assertTrue(any(("|tag" in line and line.endswith("|TAG") for line in out)))  # line 357
        _.createFile(2)  # line 358
        try:  # line 359
            sos.commit("tag")  # line 359
            _.fail()  # line 359
        except:  # line 360
            pass  # line 360
        sos.commit("tag-2", ["--tag"])  # line 361
        out = wrapChannels(lambda: sos.ls(options=["--tags"])).replace("\r", "")  # line 362
        _.assertIn("TAG tag", out)  # line 363

    def testSwitch(_):  # line 365
        _.createFile(1, "x" * 100)  # line 366
        _.createFile(2, "y")  # line 367
        sos.offline("test")  # file1-2  in initial branch commit  # line 368
        sos.branch("second")  # file1-2  switch, having same files  # line 369
        sos.switch("0")  # no change  switch back, no problem  # line 370
        sos.switch("second")  # no change  # switch back, no problem  # line 371
        _.createFile(3, "y")  # generate a file  # line 372
        try:  # uncommited changes detected  # line 373
            sos.switch("test")  # uncommited changes detected  # line 373
            _.fail()  # uncommited changes detected  # line 373
        except SystemExit as E:  # line 374
            _.assertEqual(1, E.code)  # line 374
        sos.commit("Finish")  # file1-3  commit third file into branch second  # line 375
        sos.changes()  # line 376
        sos.switch("test")  # file1-2, remove file3 from file tree  # line 377
        _.assertFalse(_.existsFile(3))  # removed when switching back to test  # line 378
        _.createFile("XXX")  # line 379
        out = wrapChannels(lambda: sos.status()).replace("\r", "")  # line 380
        _.assertIn("File tree has changes", out)  # line 381
        _.assertNotIn("File tree is unchanged", out)  # line 382
        _.assertIn("  * b00   'test'", out)  # line 383
        _.assertIn("    b01 'second'", out)  # line 384
        _.assertIn("(dirty)", out)  # one branch has commits  # line 385
        _.assertIn("(in sync)", out)  # the other doesn't  # line 386
        _.createFile(4, "xy")  # generate a file  # line 387
        sos.switch("second", ["--force"])  # avoids warning on uncommited changes, but keeps file4  # line 388
        _.assertFalse(_.existsFile(4))  # removed when forcedly switching back to test  # line 389
        _.assertTrue(_.existsFile(3))  # was restored from branch's revision r1  # line 390
        os.unlink("." + os.sep + "file1")  # remove old file1  # line 391
        sos.switch("test", ["--force"])  # should restore file1 and remove file3  # line 392
        _.assertTrue(_.existsFile(1))  # was restored from branch's revision r1  # line 393
        _.assertFalse(_.existsFile(3))  # was restored from branch's revision r1  # line 394

    def testAutoDetectVCS(_):  # line 396
        os.mkdir(".git")  # line 397
        sos.offline(sos.vcsBranches[sos.findSosVcsBase()[2]])  # create initial branch  # line 398
        with open(sos.metaFolder + os.sep + sos.metaFile, "r") as fd:  # line 399
            meta = fd.read()  # line 399
        _.assertTrue("\"master\"" in meta)  # line 400
        os.rmdir(".git")  # line 401

    def testUpdate(_):  # line 403
        sos.offline("trunk")  # create initial branch b0/r0  # line 404
        _.createFile(1, "x" * 100)  # line 405
        sos.commit("second")  # create b0/r1  # line 406

        sos.switch("/0")  # go back to b0/r0 - deletes file1  # line 408
        _.assertFalse(_.existsFile(1))  # line 409

        sos.update("/1")  # recreate file1  # line 411
        _.assertTrue(_.existsFile(1))  # line 412

        sos.commit("third", ["--force"])  # force because nothing to commit. should create r2 with same contents as r1, but as differential from r1, not from r0 (= no changes in meta folder)  # line 414
        _.assertTrue(os.path.exists(sos.branchFolder(0, 2)))  # line 415
        _.assertTrue(os.path.exists(sos.branchFolder(0, 2, file=sos.metaFile)))  # line 416
        _.assertEqual(1, len(os.listdir(sos.branchFolder(0, 2))))  # only meta data file, no differential files  # line 417

        sos.update("/1")  # do nothing, as nothing has changed  # line 419
        _.assertTrue(_.existsFile(1))  # line 420

        _.createFile(2, "y" * 100)  # line 422
#    out = wrapChannels(() -> sos.branch("other"))  # won't comply as there are changes
#    _.assertIn("--force", out)
        sos.branch("other", ["--force"])  # automatically including file 2 (as we are in simple mode)  # line 425
        _.assertTrue(_.existsFile(2))  # line 426
        sos.update("trunk", ["--add"])  # only add stuff  # line 427
        _.assertTrue(_.existsFile(2))  # line 428
        sos.update("trunk")  # nothing to do  # line 429
        _.assertFalse(_.existsFile(2))  # removes file not present in original branch  # line 430

        theirs = b"a\nb\nc\nd\ne\nf\ng\nh\nx\nx\nj\nk"  # line 432
        _.createFile(10, theirs)  # line 433
        mine = b"a\nc\nd\ne\ng\nf\nx\nh\ny\ny\nj"  # missing "b", inserted g, modified g->x, replace x/x -> y/y, removed k  # line 434
        _.createFile(11, mine)  # line 435
        _.assertEqual((b"a\nb\nc\nd\ne\nf\ng\nh\nx\nx\nj\nk", b"\n"), sos.merge(filename="." + os.sep + "file10", intoname="." + os.sep + "file11", mergeOperation=sos.MergeOperation.BOTH))  # completely recreated other file  # line 436
        _.assertEqual((b'a\nb\nc\nd\ne\ng\nf\ng\nh\ny\ny\nx\nx\nj\nk', b"\n"), sos.merge(filename="." + os.sep + "file10", intoname="." + os.sep + "file11", mergeOperation=sos.MergeOperation.INSERT))  # line 437

    def testUpdate2(_):  # line 439
        _.createFile("test.txt", "x" * 10)  # line 440
        sos.offline("trunk", ["--strict"])  # use strict mode, as timestamp differences are too small for testing  # line 441
        sos.branch("mod")  # line 442
        _.createFile("test.txt", "x" * 5 + "y" * 5)  # line 443
        time.sleep(FS_PRECISION)  # line 444
        sos.commit("mod")  # create b0/r1  # line 445
        sos.switch("trunk", ["--force"])  # should replace contents, force in case some other files were modified (e.g. during working on the code) TODO investigate more  # line 446
        with open("test.txt", "rb") as fd:  # line 447
            _.assertEqual(b"x" * 10, fd.read())  # line 447
        sos.update("mod")  # integrate changes TODO same with ask -> theirs  # line 448
        with open("test.txt", "rb") as fd:  # line 449
            _.assertEqual(b"x" * 5 + b"y" * 5, fd.read())  # line 449
        _.createFile("test.txt", "x" * 10)  # line 450
        mockInput(["t"], lambda: sos.update("mod", ["--ask-lines"]))  # line 451
        with open("test.txt", "rb") as fd:  # line 452
            _.assertEqual(b"x" * 5 + b"y" * 5, fd.read())  # line 452
        _.createFile("test.txt", "x" * 5 + "z" + "y" * 4)  # line 453
        sos.update("mod")  # auto-insert/removes (no intra-line conflict)  # line 454
        _.createFile("test.txt", "x" * 5 + "z" + "y" * 4)  # line 455
        mockInput(["t"], lambda: sos.update("mod", ["--ask"]))  # same as above with interaction -> use theirs (overwrite current file state)  # line 456
        with open("test.txt", "rb") as fd:  # line 457
            _.assertEqual(b"x" * 5 + b"y" * 5, fd.read())  # line 457

    def testIsTextType(_):  # line 459
        m = sos.Metadata(".")  # line 460
        m.c.texttype = ["*.x", "*.md", "*.md.*"]  # line 461
        m.c.bintype = ["*.md.confluence"]  # line 462
        _.assertTrue(m.isTextType("ab.txt"))  # line 463
        _.assertTrue(m.isTextType("./ab.txt"))  # line 464
        _.assertTrue(m.isTextType("bc/ab.txt"))  # line 465
        _.assertFalse(m.isTextType("bc/ab."))  # line 466
        _.assertTrue(m.isTextType("23_3.x.x"))  # line 467
        _.assertTrue(m.isTextType("dfg/dfglkjdf7/test.md"))  # line 468
        _.assertTrue(m.isTextType("./test.md.pdf"))  # line 469
        _.assertFalse(m.isTextType("./test_a.md.confluence"))  # line 470

    def testEolDet(_):  # line 472
        ''' Check correct end-of-line detection. '''  # line 473
        _.assertEqual(b"\n", sos.eoldet(b"a\nb"))  # line 474
        _.assertEqual(b"\r\n", sos.eoldet(b"a\r\nb\r\n"))  # line 475
        _.assertEqual(b"\r", sos.eoldet(b"\ra\rb"))  # line 476
        _.assertAllIn(["Inconsistent", "with "], wrapChannels(lambda: _.assertEqual(b"\n", sos.eoldet(b"\r\na\r\nb\n"))))  # line 477
        _.assertAllIn(["Inconsistent", "without"], wrapChannels(lambda: _.assertEqual(b"\n", sos.eoldet(b"\ra\nnb\n"))))  # line 478
        _.assertIsNone(sos.eoldet(b""))  # line 479
        _.assertIsNone(sos.eoldet(b"sdf"))  # line 480

    def testMerge(_):  # line 482
        ''' Check merge results depending on user options. '''  # line 483
        a = b"a\nb\ncc\nd"  # type: bytes  # line 484
        b = b"a\nb\nee\nd"  # type: bytes  # replaces cc by ee  # line 485
        _.assertEqual(b"a\nb\ncc\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.INSERT)[0])  # one-line block replacement using lineMerge  # line 486
        _.assertEqual(b"a\nb\neecc\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.INSERT, charMergeOperation=sos.MergeOperation.INSERT)[0])  # means insert changes from a into b, but don't replace  # line 487
        _.assertEqual(b"a\nb\n\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.INSERT, charMergeOperation=sos.MergeOperation.REMOVE)[0])  # means insert changes from a into b, but don't replace  # line 488
        _.assertEqual(b"a\nb\ncc\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.REMOVE)[0])  # one-line block replacement using lineMerge  # line 489
        _.assertEqual(b"a\nb\n\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.REMOVE, charMergeOperation=sos.MergeOperation.REMOVE)[0])  # line 490
        _.assertEqual(a, sos.merge(a, b, mergeOperation=sos.MergeOperation.BOTH)[0])  # keeps any changes in b  # line 491
        a = b"a\nb\ncc\nd"  # line 492
        b = b"a\nb\nee\nf\nd"  # replaces cc by block of two lines ee, f  # line 493
        _.assertEqual(b"a\nb\nee\nf\ncc\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.INSERT)[0])  # multi-line block replacement  # line 494
        _.assertEqual(b"a\nb\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.REMOVE)[0])  # line 495
        _.assertEqual(a, sos.merge(a, b, mergeOperation=sos.MergeOperation.BOTH)[0])  # keeps any changes in b  # line 496
# Test with change + insert
        _.assertEqual(b"a\nb fdcd d\ne", sos.merge(b"a\nb cd d\ne", b"a\nb fdd d\ne", charMergeOperation=sos.MergeOperation.INSERT)[0])  # line 498
        _.assertEqual(b"a\nb d d\ne", sos.merge(b"a\nb cd d\ne", b"a\nb fdd d\ne", charMergeOperation=sos.MergeOperation.REMOVE)[0])  # line 499
# Test interactive merge
        a = b"a\nb\nb\ne"  # block-wise replacement  # line 501
        b = b"a\nc\ne"  # line 502
        _.assertEqual(b, mockInput(["i"], lambda: sos.merge(a, b, mergeOperation=sos.MergeOperation.ASK)[0]))  # line 503
        _.assertEqual(a, mockInput(["t"], lambda: sos.merge(a, b, mergeOperation=sos.MergeOperation.ASK)[0]))  # line 504
        a = b"a\nb\ne"  # intra-line merge  # line 505
        _.assertEqual(b, mockInput(["i"], lambda: sos.merge(a, b, charMergeOperation=sos.MergeOperation.ASK)[0]))  # line 506
        _.assertEqual(a, mockInput(["t"], lambda: sos.merge(a, b, charMergeOperation=sos.MergeOperation.ASK)[0]))  # line 507

    def testMergeEol(_):  # line 509
        _.assertEqual(b"\r\n", sos.merge(b"a\nb", b"a\r\nb")[1])  # line 510
        _.assertIn("Differing EOL-styles", wrapChannels(lambda: sos.merge(b"a\nb", b"a\r\nb")))  # expects a warning  # line 511
        _.assertIn(b"a\r\nb", sos.merge(b"a\nb", b"a\r\nb")[0])  # when in doubt, use "mine" CR-LF  # line 512
        _.assertIn(b"a\nb", sos.merge(b"a\nb", b"a\r\nb", eol=True)[0])  # line 513
        _.assertEqual(b"\n", sos.merge(b"a\nb", b"a\r\nb", eol=True)[1])  # line 514

    def testPickyMode(_):  # line 516
        ''' Confirm that picky mode reset tracked patterns after commits. '''  # line 517
        sos.offline("trunk", ["--picky"])  # line 518
        changes = sos.changes()  # line 519
        _.assertEqual(0, len(changes.additions))  # do not list any existing file as an addition  # line 520
        sos.add(".", "./file?", ["--force"])  # line 521
        _.createFile(1, "aa")  # line 522
        sos.commit("First")  # add one file  # line 523
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # line 524
        _.createFile(2, "b")  # line 525
        try:  # add nothing, because picky  # line 526
            sos.commit("Second")  # add nothing, because picky  # line 526
        except:  # line 527
            pass  # line 527
        sos.add(".", "./file?")  # line 528
        sos.commit("Third")  # line 529
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 2))))  # line 530
        out = wrapChannels(lambda: sos.log([])).replace("\r", "")  # line 531
        _.assertIn("  * r2", out)  # line 532
        _.createFile(3, prefix="sub")  # line 533
        sos.add("sub", "sub/file?")  # line 534
        changes = sos.changes()  # line 535
        _.assertEqual(1, len(changes.additions))  # line 536
        _.assertTrue("sub/file3" in changes.additions)  # line 537

    def testTrackedSubfolder(_):  # line 539
        ''' See if patterns for files in sub folders are picked up correctly. '''  # line 540
        os.mkdir("." + os.sep + "sub")  # line 541
        sos.offline("trunk", ["--track"])  # line 542
        _.createFile(1, "x")  # line 543
        _.createFile(1, "x", prefix="sub")  # line 544
        sos.add(".", "./file?")  # add glob pattern to track  # line 545
        sos.commit("First")  # line 546
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # one new file + meta file  # line 547
        sos.add(".", "sub/file?")  # add glob pattern to track  # line 548
        sos.commit("Second")  # one new file + meta  # line 549
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # one new file + meta file  # line 550
        os.unlink("file1")  # remove from basefolder  # line 551
        _.createFile(2, "y")  # line 552
        sos.remove(".", "sub/file?")  # line 553
        try:  # raises Exit. TODO test the "suggest a pattern" case  # line 554
            sos.remove(".", "sub/bla")  # raises Exit. TODO test the "suggest a pattern" case  # line 554
            _.fail()  # raises Exit. TODO test the "suggest a pattern" case  # line 554
        except:  # line 555
            pass  # line 555
        sos.commit("Third")  # line 556
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 2))))  # one new file + meta  # line 557
# TODO also check if /file1 and sub/file1 were removed from index

    def testTrackedMode(_):  # line 560
        ''' Difference in semantics vs simple mode:
          - For remote/other branch we can only know and consider tracked files, thus ignoring all complexity stemming from handling addition of untracked files.
          - For current branch, we can take into account tracked and untracked ones, in theory, but it doesn't make sense.
        In conclusion, using the union of tracking patterns from both sides to find affected files makes sense, but disallow deleting files not present in remote branch.
    '''  # line 565
        sos.offline("test", options=["--track"])  # set up repo in tracking mode (SVN- or gitless-style)  # line 566
        _.createFile(1)  # line 567
        _.createFile("a123a")  # untracked file "a123a"  # line 568
        sos.add(".", "./file?")  # add glob tracking pattern  # line 569
        sos.commit("second")  # versions "file1"  # line 570
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # one new file + meta file  # line 571
        out = wrapChannels(lambda: sos.status()).replace("\r", "")  # line 572
        _.assertIn("  | ./file?", out)  # line 573

        _.createFile(2)  # untracked file "file2"  # line 575
        sos.commit("third")  # versions "file2"  # line 576
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 2))))  # one new file + meta file  # line 577

        os.mkdir("." + os.sep + "sub")  # line 579
        _.createFile(3, prefix="sub")  # untracked file "sub/file3"  # line 580
        sos.commit("fourth", ["--force"])  # no tracking pattern matches the subfolder  # line 581
        _.assertEqual(1, len(os.listdir(sos.branchFolder(0, 3))))  # meta file only, no other tracked path/file  # line 582

        sos.branch("Other")  # second branch containing file1 and file2 tracked by "./file?"  # line 584
        sos.remove(".", "./file?")  # remove tracking pattern, but don't touch previously created and versioned files  # line 585
        sos.add(".", "./a*a")  # add tracking pattern  # line 586
        changes = sos.changes()  # should pick up addition only, because tracked, but not the deletion, as not tracked anymore  # line 587
        _.assertEqual(0, len(changes.modifications))  # line 588
        _.assertEqual(0, len(changes.deletions))  # not tracked anymore, but contained in version history and not removed  # line 589
        _.assertEqual(1, len(changes.additions))  # detected one addition "a123a", but won't recognize untracking files as deletion  # line 590

        sos.commit("Second_2")  # line 592
        _.assertEqual(2, len(os.listdir(sos.branchFolder(1, 1))))  # "a123a" + meta file  # line 593
        _.existsFile(1, b"x" * 10)  # line 594
        _.existsFile(2, b"x" * 10)  # line 595

        sos.switch("test")  # go back to first branch - tracks only "file?", but not "a*a"  # line 597
        _.existsFile(1, b"x" * 10)  # line 598
        _.existsFile("a123a", b"x" * 10)  # line 599

        sos.update("Other")  # integrate tracked files and tracking pattern from second branch into working state of master branch  # line 601
        _.assertTrue(os.path.exists("." + os.sep + "file1"))  # line 602
        _.assertTrue(os.path.exists("." + os.sep + "a123a"))  # line 603

        _.createFile("axxxa")  # new file that should be tracked on "test" now that we integrated "Other"  # line 605
        sos.commit("fifth")  # create new revision after integrating updates from second branch  # line 606
        _.assertEqual(3, len(os.listdir(sos.branchFolder(0, 4))))  # one new file from other branch + one new in current folder + meta file  # line 607
        sos.switch("Other")  # switch back to just integrated branch that tracks only "a*a" - shouldn't do anything  # line 608
        _.assertTrue(os.path.exists("." + os.sep + "file1"))  # line 609
        _.assertTrue(os.path.exists("." + os.sep + "a123a"))  # line 610
        _.assertFalse(os.path.exists("." + os.sep + "axxxa"))  # because tracked in both branches, but not present in other -> delete in file tree  # line 611
# TODO test switch --meta

    def testLsTracked(_):  # line 614
        sos.offline("test", options=["--track"])  # set up repo in tracking mode (SVN- or gitless-style)  # line 615
        _.createFile(1)  # line 616
        _.createFile("foo")  # line 617
        sos.add(".", "./file*")  # capture one file  # line 618
        sos.ls()  # line 619
        out = sos.safeSplit(wrapChannels(lambda: sos.ls()).replace("\r", ""), "\n")  # line 620
        _.assertInAny("TRK file1  (file*)", out)  # line 621
        _.assertNotInAny("... file1  (file*)", out)  # line 622
        _.assertInAny("··· foo", out)  # line 623
        out = sos.safeSplit(wrapChannels(lambda: sos.ls(options=["--patterns"])).replace("\r", ""), "\n")  # line 624
        _.assertInAny("TRK file*", out)  # line 625
        _.createFile("a", prefix="sub")  # line 626
        sos.add("sub", "sub/a")  # line 627
        sos.ls("sub")  # line 628
        _.assertIn("TRK a  (a)", sos.safeSplit(wrapChannels(lambda: sos.ls("sub")).replace("\r", ""), "\n"))  # line 629

    def testLineMerge(_):  # line 631
        _.assertEqual("xabc", sos.lineMerge("xabc", "a bd"))  # line 632
        _.assertEqual("xabxxc", sos.lineMerge("xabxxc", "a bd"))  # line 633
        _.assertEqual("xa bdc", sos.lineMerge("xabc", "a bd", mergeOperation=sos.MergeOperation.INSERT))  # line 634
        _.assertEqual("ab", sos.lineMerge("xabc", "a bd", mergeOperation=sos.MergeOperation.REMOVE))  # line 635

    def testCompression(_):  # TODO test output ratio/advantage, also depending on compress flag set or not  # line 637
        _.createFile(1)  # line 638
        sos.offline("master", options=["--force"])  # line 639
        out = wrapChannels(lambda: sos.changes(options=['--progress'])).replace("\r", "").split("\n")  # line 640
        _.assertFalse(any(("Compression advantage" in line for line in out)))  # simple mode should always print this to stdout  # line 641
        _.assertTrue(_.existsFile(sos.branchFolder(0, 0, file="b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa"), b"x" * 10))  # line 642
        setRepoFlag("compress", True)  # was plain = uncompressed before  # line 643
        _.createFile(2)  # line 644
        out = wrapChannels(lambda: sos.commit("Added file2", options=['--progress'])).replace("\r", "").split("\n")  # line 645
        _.assertTrue(any(("Compression advantage" in line for line in out)))  # line 646
        _.assertTrue(_.existsFile(sos.branchFolder(0, 1, file="03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2")))  # exists  # line 647
        _.assertFalse(_.existsFile(sos.branchFolder(0, 1, file="03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2"), b"x" * 10))  # but is compressed instead  # line 648

    def testLocalConfig(_):  # line 650
        sos.offline("bla", options=[])  # line 651
        try:  # line 652
            sos.config(["set", "ignores", "one;two"], options=["--local"])  # line 652
        except SystemExit as E:  # line 653
            _.assertEqual(0, E.code)  # line 653
        _.assertTrue(checkRepoFlag("ignores", value=["one", "two"]))  # line 654

    def testConfigVariations(_):  # line 656
        def makeRepo():  # line 657
            try:  # line 658
                os.unlink("file1")  # line 658
            except:  # line 659
                pass  # line 659
            sos.offline("master", options=["--force"])  # line 660
            _.createFile(1)  # line 661
            sos.commit("Added file1")  # line 662
        try:  # line 663
            sos.config(["set", "strict", "on"])  # line 663
        except SystemExit as E:  # line 664
            _.assertEqual(0, E.code)  # line 664
        makeRepo()  # line 665
        _.assertTrue(checkRepoFlag("strict", True))  # line 666
        try:  # line 667
            sos.config(["set", "strict", "off"])  # line 667
        except SystemExit as E:  # line 668
            _.assertEqual(0, E.code)  # line 668
        makeRepo()  # line 669
        _.assertTrue(checkRepoFlag("strict", False))  # line 670
        try:  # line 671
            sos.config(["set", "strict", "yes"])  # line 671
        except SystemExit as E:  # line 672
            _.assertEqual(0, E.code)  # line 672
        makeRepo()  # line 673
        _.assertTrue(checkRepoFlag("strict", True))  # line 674
        try:  # line 675
            sos.config(["set", "strict", "no"])  # line 675
        except SystemExit as E:  # line 676
            _.assertEqual(0, E.code)  # line 676
        makeRepo()  # line 677
        _.assertTrue(checkRepoFlag("strict", False))  # line 678
        try:  # line 679
            sos.config(["set", "strict", "1"])  # line 679
        except SystemExit as E:  # line 680
            _.assertEqual(0, E.code)  # line 680
        makeRepo()  # line 681
        _.assertTrue(checkRepoFlag("strict", True))  # line 682
        try:  # line 683
            sos.config(["set", "strict", "0"])  # line 683
        except SystemExit as E:  # line 684
            _.assertEqual(0, E.code)  # line 684
        makeRepo()  # line 685
        _.assertTrue(checkRepoFlag("strict", False))  # line 686
        try:  # line 687
            sos.config(["set", "strict", "true"])  # line 687
        except SystemExit as E:  # line 688
            _.assertEqual(0, E.code)  # line 688
        makeRepo()  # line 689
        _.assertTrue(checkRepoFlag("strict", True))  # line 690
        try:  # line 691
            sos.config(["set", "strict", "false"])  # line 691
        except SystemExit as E:  # line 692
            _.assertEqual(0, E.code)  # line 692
        makeRepo()  # line 693
        _.assertTrue(checkRepoFlag("strict", False))  # line 694
        try:  # line 695
            sos.config(["set", "strict", "enable"])  # line 695
        except SystemExit as E:  # line 696
            _.assertEqual(0, E.code)  # line 696
        makeRepo()  # line 697
        _.assertTrue(checkRepoFlag("strict", True))  # line 698
        try:  # line 699
            sos.config(["set", "strict", "disable"])  # line 699
        except SystemExit as E:  # line 700
            _.assertEqual(0, E.code)  # line 700
        makeRepo()  # line 701
        _.assertTrue(checkRepoFlag("strict", False))  # line 702
        try:  # line 703
            sos.config(["set", "strict", "enabled"])  # line 703
        except SystemExit as E:  # line 704
            _.assertEqual(0, E.code)  # line 704
        makeRepo()  # line 705
        _.assertTrue(checkRepoFlag("strict", True))  # line 706
        try:  # line 707
            sos.config(["set", "strict", "disabled"])  # line 707
        except SystemExit as E:  # line 708
            _.assertEqual(0, E.code)  # line 708
        makeRepo()  # line 709
        _.assertTrue(checkRepoFlag("strict", False))  # line 710
        try:  # line 711
            sos.config(["set", "strict", "nope"])  # line 711
            _.fail()  # line 711
        except SystemExit as E:  # line 712
            _.assertEqual(1, E.code)  # line 712

    def testLsSimple(_):  # line 714
        _.createFile(1)  # line 715
        _.createFile("foo")  # line 716
        _.createFile("ign1")  # line 717
        _.createFile("ign2")  # line 718
        sos.offline("test")  # set up repo in tracking mode (SVN- or gitless-style)  # line 719
        try:  # define an ignore pattern  # line 720
            sos.config(["set", "ignores", "ign1"])  # define an ignore pattern  # line 720
        except SystemExit as E:  # line 721
            _.assertEqual(0, E.code)  # line 721
        try:  # additional ignore pattern  # line 722
            sos.config(["add", "ignores", "ign2"])  # additional ignore pattern  # line 722
        except SystemExit as E:  # line 723
            _.assertEqual(0, E.code)  # line 723
        try:  # define a list of ignore patterns  # line 724
            sos.config(["set", "ignoresWhitelist", "ign1;ign2"])  # define a list of ignore patterns  # line 724
        except SystemExit as E:  # line 725
            _.assertEqual(0, E.code)  # line 725
        out = wrapChannels(lambda: sos.config(["show"])).replace("\r", "")  # line 726
        _.assertIn("             ignores [global]  ['ign1', 'ign2']", out)  # line 727
        out = sos.safeSplit(wrapChannels(lambda: sos.ls()).replace("\r", ""), "\n")  # line 728
        _.assertInAny('··· file1', out)  # line 729
        _.assertInAny('··· ign1', out)  # line 730
        _.assertInAny('··· ign2', out)  # line 731
        try:  # line 732
            sos.config(["rm", "foo", "bar"])  # line 732
            _.fail()  # line 732
        except SystemExit as E:  # line 733
            _.assertEqual(1, E.code)  # line 733
        try:  # line 734
            sos.config(["rm", "ignores", "foo"])  # line 734
            _.fail()  # line 734
        except SystemExit as E:  # line 735
            _.assertEqual(1, E.code)  # line 735
        try:  # line 736
            sos.config(["rm", "ignores", "ign1"])  # line 736
        except SystemExit as E:  # line 737
            _.assertEqual(0, E.code)  # line 737
        try:  # remove ignore pattern  # line 738
            sos.config(["unset", "ignoresWhitelist"])  # remove ignore pattern  # line 738
        except SystemExit as E:  # line 739
            _.assertEqual(0, E.code)  # line 739
        out = sos.safeSplit(wrapChannels(lambda: sos.ls()).replace("\r", ""), "\n")  # line 740
        _.assertInAny('··· ign1', out)  # line 741
        _.assertInAny('IGN ign2', out)  # line 742
        _.assertNotInAny('··· ign2', out)  # line 743

    def testWhitelist(_):  # line 745
# TODO test same for simple mode
        _.createFile(1)  # line 747
        sos.defaults.ignores[:] = ["file*"]  # replace in-place  # line 748
        sos.offline("xx", ["--track", "--strict"])  # because nothing to commit due to ignore pattern  # line 749
        sos.add(".", "./file*")  # add tracking pattern for "file1"  # line 750
        sos.commit(options=["--force"])  # attempt to commit the file  # line 751
        _.assertEqual(1, len(os.listdir(sos.branchFolder(0, 1))))  # only meta data, file1 was ignored  # line 752
        try:  # Exit because dirty  # line 753
            sos.online()  # Exit because dirty  # line 753
            _.fail()  # Exit because dirty  # line 753
        except:  # exception expected  # line 754
            pass  # exception expected  # line 754
        _.createFile("x2")  # add another change  # line 755
        sos.add(".", "./x?")  # add tracking pattern for "file1"  # line 756
        try:  # force beyond dirty flag check  # line 757
            sos.online(["--force"])  # force beyond dirty flag check  # line 757
            _.fail()  # force beyond dirty flag check  # line 757
        except:  # line 758
            pass  # line 758
        sos.online(["--force", "--force"])  # force beyond file tree modifications check  # line 759
        _.assertFalse(os.path.exists(sos.metaFolder))  # line 760

        _.createFile(1)  # line 762
        sos.defaults.ignoresWhitelist[:] = ["file*"]  # line 763
        sos.offline("xx", ["--track"])  # line 764
        sos.add(".", "./file*")  # line 765
        sos.commit()  # should NOT ask for force here  # line 766
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # meta data and "file1", file1 was whitelisted  # line 767

    def testRemove(_):  # line 769
        _.createFile(1, "x" * 100)  # line 770
        sos.offline("trunk")  # line 771
        try:  # line 772
            sos.delete("trunk")  # line 772
            _fail()  # line 772
        except:  # line 773
            pass  # line 773
        _.createFile(2, "y" * 10)  # line 774
        sos.branch("added")  # creates new branch, writes repo metadata, and therefore creates backup copy  # line 775
        sos.delete("trunk")  # line 776
        _.assertEqual(3, len(os.listdir("." + os.sep + sos.metaFolder)))  # meta data file, backup and "b1"  # line 777
        _.assertTrue(os.path.exists("." + os.sep + sos.metaFolder + os.sep + "b1"))  # line 778
        _.assertFalse(os.path.exists("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 779
        sos.branch("next")  # line 780
        _.createFile(3, "y" * 10)  # make a change  # line 781
        sos.delete("added", "--force")  # should succeed  # line 782

    def testUsage(_):  # line 784
        try:  # TODO expect sys.exit(0)  # line 785
            sos.usage()  # TODO expect sys.exit(0)  # line 785
            _.fail()  # TODO expect sys.exit(0)  # line 785
        except:  # line 786
            pass  # line 786
        try:  # line 787
            sos.usage(short=True)  # line 787
            _.fail()  # line 787
        except:  # line 788
            pass  # line 788

    def testOnlyExcept(_):  # line 790
        ''' Test blacklist glob rules. '''  # line 791
        sos.offline(options=["--track"])  # line 792
        _.createFile("a.1")  # line 793
        _.createFile("a.2")  # line 794
        _.createFile("b.1")  # line 795
        _.createFile("b.2")  # line 796
        sos.add(".", "./a.?")  # line 797
        sos.add(".", "./?.1", negative=True)  # line 798
        out = wrapChannels(lambda: sos.commit())  # line 799
        _.assertIn("ADD ./a.2", out)  # line 800
        _.assertNotIn("ADD ./a.1", out)  # line 801
        _.assertNotIn("ADD ./b.1", out)  # line 802
        _.assertNotIn("ADD ./b.2", out)  # line 803

    def testOnly(_):  # line 805
        _.assertEqual((_coconut.frozenset(("./A", "x/B")), _coconut.frozenset(("./C",))), sos.parseOnlyOptions(".", ["abc", "def", "--only", "A", "--x", "--only", "x/B", "--except", "C", "--only"]))  # line 806
        _.assertEqual(_coconut.frozenset(("B",)), sos.conditionalIntersection(_coconut.frozenset(("A", "B", "C")), _coconut.frozenset(("B", "D"))))  # line 807
        _.assertEqual(_coconut.frozenset(("B", "D")), sos.conditionalIntersection(_coconut.frozenset(), _coconut.frozenset(("B", "D"))))  # line 808
        _.assertEqual(_coconut.frozenset(("B", "D")), sos.conditionalIntersection(None, _coconut.frozenset(("B", "D"))))  # line 809
        sos.offline(options=["--track", "--strict"])  # line 810
        _.createFile(1)  # line 811
        _.createFile(2)  # line 812
        sos.add(".", "./file1")  # line 813
        sos.add(".", "./file2")  # line 814
        sos.commit(onlys=_coconut.frozenset(("./file1",)))  # line 815
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # only meta and file1  # line 816
        sos.commit()  # adds also file2  # line 817
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 2))))  # only meta and file1  # line 818
        _.createFile(1, "cc")  # modify both files  # line 819
        _.createFile(2, "dd")  # line 820
        try:  # line 821
            sos.config(["set", "texttype", "file2"])  # line 821
        except SystemExit as E:  # line 822
            _.assertEqual(0, E.code)  # line 822
        changes = sos.changes(excps=_coconut.frozenset(("./file1",)))  # line 823
        _.assertEqual(1, len(changes.modifications))  # only file2  # line 824
        _.assertTrue("./file2" in changes.modifications)  # line 825
        _.assertAllIn(["DIF ./file2", "<No newline>"], wrapChannels(lambda: sos.diff(onlys=_coconut.frozenset(("./file2",)))))  # line 826
        _.assertAllNotIn(["MOD ./file1", "DIF ./file1", "MOD ./file2"], wrapChannels(lambda: sos.diff(onlys=_coconut.frozenset(("./file2",)))))  # line 827

    def testDiff(_):  # line 829
        try:  # manually mark this file as "textual"  # line 830
            sos.config(["set", "texttype", "file1"])  # manually mark this file as "textual"  # line 830
        except SystemExit as E:  # line 831
            _.assertEqual(0, E.code)  # line 831
        sos.offline(options=["--strict"])  # line 832
        _.createFile(1)  # line 833
        _.createFile(2)  # line 834
        sos.commit()  # line 835
        _.createFile(1, "sdfsdgfsdf")  # line 836
        _.createFile(2, "12343")  # line 837
        sos.commit()  # line 838
        _.createFile(1, "foobar")  # line 839
        _.createFile(3)  # line 840
        out = wrapChannels(lambda: sos.diff("/-2"))  # compare with r1 (second counting from last which is r2)  # line 841
        _.assertIn("ADD ./file3", out)  # line 842
        _.assertAllIn(["MOD ./file2", "DIF ./file1", "- | 0001 |xxxxxxxxxx|", "+ | 0000 |foobar|"], out)  # line 843
        _.assertAllNotIn(["MOD ./file1", "DIF ./file1"], wrapChannels(lambda: sos.diff("/-2", onlys=_coconut.frozenset(("./file2",)))))  # line 844

    def testReorderRenameActions(_):  # line 846
        result = sos.reorderRenameActions([("123", "312"), ("312", "132"), ("321", "123")], exitOnConflict=False)  # type: Tuple[str, str]  # line 847
        _.assertEqual([("312", "132"), ("123", "312"), ("321", "123")], result)  # line 848
        try:  # line 849
            sos.reorderRenameActions([("123", "312"), ("312", "123")], exitOnConflict=True)  # line 849
            _.fail()  # line 849
        except:  # line 850
            pass  # line 850

    def testMove(_):  # line 852
        sos.offline(options=["--strict", "--track"])  # line 853
        _.createFile(1)  # line 854
        sos.add(".", "./file?")  # line 855
# test source folder missing
        try:  # line 857
            sos.move("sub", "sub/file?", ".", "?file")  # line 857
            _.fail()  # line 857
        except:  # line 858
            pass  # line 858
# test target folder missing: create it
        sos.move(".", "./file?", "sub", "sub/file?")  # line 860
        _.assertTrue(os.path.exists("sub"))  # line 861
        _.assertTrue(os.path.exists("sub/file1"))  # line 862
        _.assertFalse(os.path.exists("file1"))  # line 863
# test move
        sos.move("sub", "sub/file?", ".", "./?file")  # line 865
        _.assertTrue(os.path.exists("1file"))  # line 866
        _.assertFalse(os.path.exists("sub/file1"))  # line 867
# test nothing matches source pattern
        try:  # line 869
            sos.move(".", "a*", ".", "b*")  # line 869
            _.fail()  # line 869
        except:  # line 870
            pass  # line 870
        sos.add(".", "*")  # anything pattern  # line 871
        try:  # TODO check that alternative pattern "*" was suggested (1 hit)  # line 872
            sos.move(".", "a*", ".", "b*")  # TODO check that alternative pattern "*" was suggested (1 hit)  # line 872
            _.fail()  # TODO check that alternative pattern "*" was suggested (1 hit)  # line 872
        except:  # line 873
            pass  # line 873
# test rename no conflict
        _.createFile(1)  # line 875
        _.createFile(2)  # line 876
        _.createFile(3)  # line 877
        sos.add(".", "./file*")  # line 878
        try:  # define an ignore pattern  # line 879
            sos.config(["set", "ignores", "file3;file4"])  # define an ignore pattern  # line 879
        except SystemExit as E:  # line 880
            _.assertEqual(0, E.code)  # line 880
        try:  # line 881
            sos.config(["set", "ignoresWhitelist", "file3"])  # line 881
        except SystemExit as E:  # line 882
            _.assertEqual(0, E.code)  # line 882
        sos.move(".", "./file*", ".", "fi*le")  # line 883
        _.assertTrue(all((os.path.exists("fi%dle" % i) for i in range(1, 4))))  # line 884
        _.assertFalse(os.path.exists("fi4le"))  # line 885
# test rename solvable conflicts
        [_.createFile("%s-%s-%s" % tuple((c for c in n))) for n in ["312", "321", "123", "231"]]  # line 887
#    sos.move("?-?-?")
# test rename unsolvable conflicts
# test --soft option
        sos.remove(".", "./?file")  # was renamed before  # line 891
        sos.add(".", "./?a?b", ["--force"])  # line 892
        sos.move(".", "./?a?b", ".", "./a?b?", ["--force", "--soft"])  # line 893
        _.createFile("1a2b")  # should not be tracked  # line 894
        _.createFile("a1b2")  # should be tracked  # line 895
        sos.commit()  # line 896
        _.assertEqual(2, len(os.listdir(sos.branchFolder(0, 1))))  # line 897
        _.assertTrue(os.path.exists(sos.branchFolder(0, 1, file="93b38f90892eb5c57779ca9c0b6fbdf6774daeee3342f56f3e78eb2fe5336c50")))  # a1b2  # line 898
        _.createFile("1a1b1")  # line 899
        _.createFile("1a1b2")  # line 900
        sos.add(".", "?a?b*")  # line 901
        _.assertIn("not unique", wrapChannels(lambda: sos.move(".", "?a?b*", ".", "z?z?")))  # should raise error due to same target name  # line 902
# TODO only rename if actually any files are versioned? or simply what is alife?
# TODO add test if two single question marks will be moved into adjacent characters

    def testHashCollision(_):  # line 906
        sos.offline()  # line 907
        _.createFile(1)  # line 908
        os.mkdir(sos.branchFolder(0, 1))  # line 909
        _.createFile("b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa", prefix=sos.branchFolder(0, 1))  # line 910
        _.createFile(1)  # line 911
        try:  # should exit with error due to collision detection  # line 912
            sos.commit()  # should exit with error due to collision detection  # line 912
            _.fail()  # should exit with error due to collision detection  # line 912
        except SystemExit as E:  # TODO will capture exit(0) which is wrong, change to check code in all places  # line 913
            _.assertEqual(1, E.code)  # TODO will capture exit(0) which is wrong, change to check code in all places  # line 913

    def testFindBase(_):  # line 915
        old = os.getcwd()  # line 916
        try:  # line 917
            os.mkdir("." + os.sep + ".git")  # line 918
            os.makedirs("." + os.sep + "a" + os.sep + sos.metaFolder)  # line 919
            os.makedirs("." + os.sep + "a" + os.sep + "b")  # line 920
            os.chdir("a" + os.sep + "b")  # line 921
            s, vcs, cmd = sos.findSosVcsBase()  # line 922
            _.assertIsNotNone(s)  # line 923
            _.assertIsNotNone(vcs)  # line 924
            _.assertEqual("git", cmd)  # line 925
        finally:  # line 926
            os.chdir(old)  # line 926

# TODO test command line operation --sos vs. --vcs
# check exact output instead of only expected exception/fail


if __name__ == '__main__':  # line 932
    logging.basicConfig(level=logging.DEBUG if '-v' in sys.argv or "true" in [os.getenv("DEBUG", "false").strip().lower(), os.getenv("CI", "false").strip().lower()] else logging.INFO, stream=sys.stderr, format="%(asctime)-23s %(levelname)-8s %(name)s:%(lineno)d | %(message)s" if '-v' in sys.argv or os.getenv("DEBUG", "false") == "true" else "%(message)s")  # line 933
    c = configr.Configr("sos")  # line 934
    c.loadSettings()  # line 935
    if len(c.keys()) > 0:  # line 936
        sos.Exit("Cannot run test suite with existing local SOS user configuration (would affect results)")  # line 936
    unittest.main(testRunner=debugTestRunner() if '-v' in sys.argv and not os.getenv("CI", "false").lower() == "true" else None)  # warnings = "ignore")  # line 937



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0xda67bb37

# Compiled with Coconut version 1.3.1-post_dev17 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

# Copyright Arne Bachmann
# This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Utiliy functions
import bz2  # line 5
import codecs  # line 5
import difflib  # line 5
import hashlib  # line 5
import logging  # line 5
import os  # line 5
import re  # line 5
sys = _coconut_sys  # line 5
import time  # line 5
START_TIME = time.time()  # line 5
try:  # line 6
    import enum  # line 6
except:  # line 7
    raise Exception("Please install SOS via 'pip install -U sos-vcs[backport]' to get enum support for Python versions prior 3.4")  # line 7
try:  # line 8
    from typing import Any  # only required for mypy  # line 9
    from typing import Callable  # only required for mypy  # line 9
    from typing import Dict  # only required for mypy  # line 9
    from typing import FrozenSet  # only required for mypy  # line 9
    from typing import Generic  # only required for mypy  # line 9
    from typing import IO  # only required for mypy  # line 9
    from typing import List  # only required for mypy  # line 9
    from typing import Sequence  # only required for mypy  # line 9
    from typing import Set  # only required for mypy  # line 9
    from typing import Tuple  # only required for mypy  # line 9
    from typing import Type  # only required for mypy  # line 9
    from typing import TypeVar  # only required for mypy  # line 9
    from typing import Union  # only required for mypy  # line 9
    Number = TypeVar("Number", int, float)  # line 10
    DataType = TypeVar("DataType", BranchInfo, ChangeSet, MergeBlock, PathInfo)  # line 11
except:  # typing not available (prior to Python 3.5)  # line 12
    pass  # typing not available (prior to Python 3.5)  # line 12
try:  # line 13
    import wcwidth  # line 13
except:  # optional dependency  # line 14
    pass  # optional dependency  # line 14


verbose = os.environ.get("DEBUG", "False").lower() == "true" or '--verbose' in sys.argv or '-v' in sys.argv  # imported from utility, and only modified here  # line 17

# Classes
class Accessor(dict):  # line 20
    ''' Dictionary with attribute access. Writing only supported via dictionary access. '''  # line 21
    def __init__(_, mapping: 'Dict[str, Any]'):  # line 22
        dict.__init__(_, mapping)  # line 22
    @_coconut_tco  # line 23
    def __getattribute__(_, name: 'str') -> 'Any':  # line 23
        try:  # line 24
            return _[name]  # line 24
        except:  # line 25
            return _coconut_tail_call(dict.__getattribute__, _, name)  # line 25

class Counter(Generic[Number]):  # line 27
    ''' A simple counter. Can be augmented to return the last value instead. '''  # line 28
    def __init__(_, initial: 'Number'=0):  # line 29
        _.value = initial  # type: Number  # line 29
    def inc(_, by: 'Number'=1) -> 'Number':  # line 30
        _.value += by  # line 30
        return _.value  # line 30

class ProgressIndicator(Counter):  # line 32
    ''' Manages a rotating progress indicator. '''  # line 33
    def __init__(_, callback: 'Optional[_coconut.typing.Callable[[str], None]]'=None):  # line 34
        super(ProgressIndicator, _).__init__(-1)  # line 34
        _.timer = time.time()  # type: float  # line 34
        _.callback = callback  # type: Optional[_coconut.typing.Callable[[str], None]]  # line 34
    def getIndicator(_) -> '_coconut.typing.Optional[str]':  # line 35
        newtime = time.time()  # line 36
        if newtime - _.timer < .1:  # line 37
            return None  # line 37
        _.timer = newtime  # line 38
        sign = PROGRESS_MARKER[int(_.inc() % 4)]  # type: str  # line 39
        if _.callback:  # line 40
            _.callback(sign)  # line 40
        return sign  # line 41

class Logger:  # line 43
    ''' Logger that supports many items. '''  # line 44
    def __init__(_, log):  # line 45
        _._log = log  # line 45
    def debug(_, *s):  # line 46
        _._log.debug(sjoin(*s))  # line 46
    def info(_, *s):  # line 47
        _._log.info(sjoin(*s))  # line 47
    def warn(_, *s):  # line 48
        _._log.warning(sjoin(*s))  # line 48
    def error(_, *s):  # line 49
        _._log.error(sjoin(*s))  # line 49


# Constants
_log = Logger(logging.getLogger(__name__))  # line 53
debug, info, warn, error = _log.debug, _log.info, _log.warn, _log.error  # line 53
UTF8 = "utf_8"  # early used constant, not defined in standard library  # line 54
CONFIGURABLE_FLAGS = ["strict", "track", "picky", "compress"]  # line 55
CONFIGURABLE_LISTS = ["texttype", "bintype", "ignores", "ignoreDirs", "ignoresWhitelist", "ignoreDirsWhitelist"]  # line 56
GLOBAL_LISTS = ["ignores", "ignoreDirs", "ignoresWhitelist", "ignoreDirsWhitelist"]  # line 57
TRUTH_VALUES = ["true", "yes", "on", "1", "enable", "enabled"]  # all lower-case normalized  # line 58
FALSE_VALUES = ["false", "no", "off", "0", "disable", "disabled"]  # line 59
PROGRESS_MARKER = ["|", "/", "-", "\\"]  # line 60
BACKUP_SUFFIX = "_last"  # line 61
metaFolder = ".sos"  # type: str  # line 62
metaFile = ".meta"  # type: str  # line 63
metaBack = metaFile + BACKUP_SUFFIX  # type: str  # line 64
bufSize = 1 << 20  # type: int  # 1 MiB  # line 65
SVN = "svn"  # line 66
SLASH = "/"  # line 67
MEBI = 1 << 20  # line 68
vcsFolders = {".svn": SVN, ".git": "git", ".bzr": "bzr", ".hg": "hg", ".fslckout": "fossil", "_FOSSIL_": "fossil", ".CVS": "cvs"}  # type: Dict[str, str]  # line 69
vcsBranches = {SVN: "trunk", "git": "master", "bzr": "trunk", "hg": "default", "fossil": None, "cvs": None}  # type: Dict[str, _coconut.typing.Optional[str]]  # line 70
NL_NAMES = {None: "<No newline>", b"\r\n": "<CR+LF>", b"\n\r": "<LF+CR>", b"\n": "<LF>", b"\r": "<CR>"}  # type: Dict[bytes, str]  # line 71
lateBinding = Accessor({"verbose": False, "start": 0})  # type: Accessor  # line 72


# Enums
MergeOperation = enum.Enum("MergeOperation", {"INSERT": 1, "REMOVE": 2, "BOTH": 3, "ASK": 4})  # insert remote changes into current, remove remote deletions from current, do both (replicates remote state), or ask per block  # line 76
MergeBlockType = enum.Enum("MergeBlockType", "KEEP INSERT REMOVE REPLACE MOVE")  # modify = intra-line changes, replace = full block replacement  # line 77


# Value types
class BranchInfo(_coconut_NamedTuple("BranchInfo", [("number", 'int'), ("ctime", 'int'), ("name", '_coconut.typing.Optional[str]'), ("inSync", 'bool'), ("tracked", 'List[str]'), ("untracked", 'List[str]')])):  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 81
    __slots__ = ()  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 81
    __ne__ = _coconut.object.__ne__  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 81
    def __new__(_cls, number, ctime, name=None, inSync=False, tracked=[], untracked=[]):  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 81
        return _coconut.tuple.__new__(_cls, (number, ctime, name, inSync, tracked, untracked))  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 81
# tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place
class CommitInfo(_coconut_NamedTuple("CommitInfo", [("number", 'int'), ("ctime", 'int'), ("message", '_coconut.typing.Optional[str]')])):  # line 82
    __slots__ = ()  # line 82
    __ne__ = _coconut.object.__ne__  # line 82
    def __new__(_cls, number, ctime, message=None):  # line 82
        return _coconut.tuple.__new__(_cls, (number, ctime, message))  # line 82

class PathInfo(_coconut_NamedTuple("PathInfo", [("nameHash", 'str'), ("size", '_coconut.typing.Optional[int]'), ("mtime", 'int'), ("hash", '_coconut.typing.Optional[str]')])):  # size == None means deleted in this revision  # line 83
    __slots__ = ()  # size == None means deleted in this revision  # line 83
    __ne__ = _coconut.object.__ne__  # size == None means deleted in this revision  # line 83
class ChangeSet(_coconut_NamedTuple("ChangeSet", [("additions", 'Dict[str, PathInfo]'), ("deletions", 'Dict[str, PathInfo]'), ("modifications", 'Dict[str, PathInfo]')])):  # avoid default assignment of {} as it leads to runtime errors (contains data on init for unknown reason)  # line 84
    __slots__ = ()  # avoid default assignment of {} as it leads to runtime errors (contains data on init for unknown reason)  # line 84
    __ne__ = _coconut.object.__ne__  # avoid default assignment of {} as it leads to runtime errors (contains data on init for unknown reason)  # line 84
class Range(_coconut_NamedTuple("Range", [("tipe", 'MergeBlockType'), ("indexes", '_coconut.typing.Sequence[int]')])):  # MergeBlockType[1,2,4], line number, length  # TODO use enum  # line 85
    __slots__ = ()  # MergeBlockType[1,2,4], line number, length  # TODO use enum  # line 85
    __ne__ = _coconut.object.__ne__  # MergeBlockType[1,2,4], line number, length  # TODO use enum  # line 85
class MergeBlock(_coconut_NamedTuple("MergeBlock", [("tipe", 'MergeBlockType'), ("lines", 'List[str]'), ("line", 'int'), ("replaces", '_coconut.typing.Optional[MergeBlock]'), ("changes", '_coconut.typing.Optional[Range]')])):  # line 86
    __slots__ = ()  # line 86
    __ne__ = _coconut.object.__ne__  # line 86
    def __new__(_cls, tipe, lines, line, replaces=None, changes=None):  # line 86
        return _coconut.tuple.__new__(_cls, (tipe, lines, line, replaces, changes))  # line 86

class GlobBlock(_coconut_NamedTuple("GlobBlock", [("isLiteral", 'bool'), ("content", 'str'), ("index", 'int')])):  # for file pattern rename/move matching  # line 87
    __slots__ = ()  # for file pattern rename/move matching  # line 87
    __ne__ = _coconut.object.__ne__  # for file pattern rename/move matching  # line 87
class GlobBlock2(_coconut_NamedTuple("GlobBlock2", [("isLiteral", 'bool'), ("content", 'str'), ("matches", 'str')])):  # matching file pattern and input filename for translation  # line 88
    __slots__ = ()  # matching file pattern and input filename for translation  # line 88
    __ne__ = _coconut.object.__ne__  # matching file pattern and input filename for translation  # line 88


# Functions
def printo(s: 'str', nl: 'str'="\n"):  # PEP528 compatibility  # line 92
    sys.stdout.buffer.write((s + nl).encode(sys.stdout.encoding, 'backslashreplace'))  # PEP528 compatibility  # line 92
    sys.stdout.flush()  # PEP528 compatibility  # line 92
def printe(s: 'str', nl: 'str'="\n"):  # line 93
    sys.stderr.buffer.write((s + nl).encode(sys.stderr.encoding, 'backslashreplace'))  # line 93
    sys.stderr.flush()  # line 93
@_coconut_tco  # for py->os access of writing filenames  # PEP 529 compatibility  # line 94
def encode(s: 'str') -> 'bytes':  # for py->os access of writing filenames  # PEP 529 compatibility  # line 94
    return _coconut_tail_call(os.fsencode, s)  # for py->os access of writing filenames  # PEP 529 compatibility  # line 94
@_coconut_tco  # for os->py access of reading filenames  # line 95
def decode(b: 'bytes') -> 'str':  # for os->py access of reading filenames  # line 95
    return _coconut_tail_call(os.fsdecode, b)  # for os->py access of reading filenames  # line 95
try:  # line 96
    import chardet  # https://github.com/chardet/chardet  # line 97
    def detectEncoding(binary: 'bytes') -> 'str':  # line 98
        return chardet.detect(binary)["encoding"]  # line 98
except:  # line 99
    def detectEncoding(binary: 'bytes') -> 'str':  # Guess the encoding  # line 100
        ''' Fallback if chardet library missing. '''  # line 101
        try:  # line 102
            binary.decode(UTF8)  # line 102
            return UTF8  # line 102
        except UnicodeError:  # line 103
            pass  # line 103
        try:  # line 104
            binary.decode("utf_16")  # line 104
            return "utf_16"  # line 104
        except UnicodeError:  # line 105
            pass  # line 105
        try:  # line 106
            binary.decode("cp1252")  # line 106
            return "cp1252"  # line 106
        except UnicodeError:  # line 107
            pass  # line 107
        return "ascii"  # this code will never be reached, as above is an 8-bit charset that always matches  # line 108

def tryOrDefault(func: '_coconut.typing.Callable[[], Any]', default: 'Any') -> 'Any':  # line 110
    try:  # line 111
        return func()  # line 111
    except:  # line 112
        return default  # line 112

def tryOrIgnore(func: '_coconut.typing.Callable[[], Any]') -> 'None':  # handle with care!  # line 114
    try:  # line 115
        func()  # line 115
    except:  # line 116
        pass  # line 116

@_coconut_tco  # line 118
def wcswidth(string: 'str') -> 'int':  # line 118
    l = 0  # type: int  # line 119
    try:  # line 120
        l = wcwidth.wcswitdh(string)  # line 121
        return len(string) if l < 0 else l  # line 122
    finally:  # line 123
        return _coconut_tail_call(len, string)  # line 123

def removePath(key: 'str', value: 'str') -> 'str':  # line 125
    ''' Cleanup of user-specified global file patterns. '''  # line 126
    return value if value in GLOBAL_LISTS or SLASH not in value else value[value.rindex(SLASH) + 1:]  # line 127

def conditionalIntersection(a: '_coconut.typing.Optional[FrozenSet[str]]', b: 'FrozenSet[str]') -> 'FrozenSet[str]':  # Used to match only arguments, or use only stored patterns  # line 129
    return a & b if a else b  # Used to match only arguments, or use only stored patterns  # line 129

def dictUpdate(dikt: 'Dict[Any, Any]', by: 'Dict[Any, Any]') -> 'Dict[Any, Any]':  # line 131
    d = {}  # line 131
    d.update(dikt)  # line 131
    d.update(by)  # line 131
    return d  # line 131

def openIt(file: 'str', mode: 'str', compress: 'bool'=False) -> 'IO[bytes]':  # Abstraction for opening both compressed and plain files  # line 133
    return bz2.BZ2File(encode(file), mode) if compress else open(encode(file), mode + "b")  # Abstraction for opening both compressed and plain files  # line 133

def eoldet(file: 'bytes') -> '_coconut.typing.Optional[bytes]':  # line 135
    ''' Determine EOL style from a binary string. '''  # line 136
    lf = file.count(b"\n")  # type: int  # line 137
    cr = file.count(b"\r")  # type: int  # line 138
    crlf = file.count(b"\r\n")  # type: int  # line 139
    if crlf > 0:  # DOS/Windows/Symbian etc.  # line 140
        if lf != crlf or cr != crlf:  # line 141
            warn("Inconsistent CR/NL count with CR+NL. Mixed EOL style detected, may cause problems during merge")  # line 141
        return b"\r\n"  # line 142
    if lf != 0 and cr != 0:  # line 143
        warn("Inconsistent CR/NL count without CR+NL. Mixed EOL style detected, may cause problems during merge")  # line 143
    if lf > cr:  # Linux/Unix  # line 144
        return b"\n"  # Linux/Unix  # line 144
    if cr > lf:  # older 8-bit machines  # line 145
        return b"\r"  # older 8-bit machines  # line 145
    return None  # no new line contained, cannot determine  # line 146

try:  # line 148
    Splittable = TypeVar("Splittable", str, bytes)  # line 148
except:  # line 149
    pass  # line 149
def safeSplit(s: 'Splittable', d: '_coconut.typing.Optional[Splittable]'=None) -> 'List[Splittable]':  # line 150
    return s.split((("\n" if isinstance(s, str) else b"\n") if d is None else d)) if len(s) > 0 else []  # line 150

def ajoin(sep: 'str', seq: '_coconut.typing.Sequence[str]', nl: 'str'="") -> 'str':  # line 152
    return (sep + (nl + sep).join(seq)) if seq else ""  # line 152

@_coconut_tco  # line 154
def sjoin(*s: 'Tuple[Any]') -> 'str':  # line 154
    return _coconut_tail_call(" ".join, [str(e) for e in s if e != ''])  # line 154

@_coconut_tco  # line 156
def hashStr(datas: 'str') -> 'str':  # line 156
    return _coconut_tail_call(hashlib.sha256(datas.encode(UTF8)).hexdigest)  # line 156

def modified(changes: 'ChangeSet', onlyBinary: 'bool'=False) -> 'bool':  # line 158
    return len(changes.additions) > 0 or len(changes.deletions) > 0 or len(changes.modifications) > 0  # line 158

def listindex(lizt: 'Sequence[Any]', what: 'Any', index: 'int'=0) -> 'int':  # line 160
    return lizt[index:].index(what) + index  # line 160

def getTermWidth() -> 'int':  # line 162
    try:  # line 163
        import termwidth  # line 163
    except:  # line 164
        return 80  # line 164
    return termwidth.getTermWidth()[0]  # line 165

def branchFolder(branch: 'int', revision: 'int', base=".", file=None) -> 'str':  # line 167
    return os.path.join(base, metaFolder, "b%d" % branch, "r%d" % revision) + ((os.sep + file) if file else "")  # line 167

def Exit(message: 'str'="", code=1):  # line 169
    printe("[EXIT%s]" % (" %.1fs" % (time.time() - START_TIME) if verbose else "") + (" " + message + "." if message != "" else ""))  # line 169
    sys.exit(code)  # line 169

def exception(E):  # line 171
    ''' Report an exception. '''  # line 172
    printo(str(E))  # line 173
    import traceback  # line 174
    traceback.print_exc()  # line 175
    traceback.print_stack()  # line 176

def hashFile(path: 'str', compress: 'bool', saveTo: '_coconut.typing.Optional[str]'=None, callback: 'Optional[_coconut.typing.Callable[[str], None]]'=None) -> 'Tuple[str, int]':  # line 178
    ''' Calculate hash of file contents, and return compressed sized, if in write mode, or zero. '''  # line 179
    if callback:  # line 180
        indicator = ProgressIndicator(callback)  # type: ProgressIndicator  # line 180
    _hash = hashlib.sha256()  # line 181
    wsize = 0  # type: int  # line 182
    if saveTo and os.path.exists(encode(saveTo)):  # line 183
        Exit("Hash conflict. Leaving revision in inconsistent state. This should happen only once in a lifetime.")  # line 183
    to = openIt(saveTo, "w", compress) if saveTo else None  # line 184
    with open(encode(path), "rb") as fd:  # line 185
        while True:  # line 186
            buffer = fd.read(bufSize)  # type: bytes  # line 187
            _hash.update(buffer)  # line 188
            if to:  # line 189
                to.write(buffer)  # line 189
            if len(buffer) < bufSize:  # line 190
                break  # line 190
            if indicator:  # line 191
                indicator.getIndicator()  # line 191
        if to:  # line 192
            to.close()  # line 193
            wsize = os.stat(encode(saveTo)).st_size  # line 194
    return (_hash.hexdigest(), wsize)  # line 195

def getAnyOfMap(map: 'Dict[str, Any]', params: '_coconut.typing.Sequence[str]', default: 'Any'=None) -> 'Any':  # line 197
    ''' Utility. '''  # line 198
    for k, v in map.items():  # line 199
        if k in params:  # line 200
            return v  # line 200
    return default  # line 201

@_coconut_tco  # line 203
def strftime(timestamp: '_coconut.typing.Optional[int]'=None) -> 'str':  # line 203
    return _coconut_tail_call(time.strftime, "%Y-%m-%d %H:%M:%S", time.localtime(timestamp / 1000. if timestamp is not None else None))  # line 203

def detectAndLoad(filename: '_coconut.typing.Optional[str]'=None, content: '_coconut.typing.Optional[bytes]'=None) -> 'Tuple[str, bytes, _coconut.typing.Sequence[str]]':  # line 205
    encoding = None  # type: str  # line 206
    eol = None  # type: _coconut.typing.Optional[bytes]  # line 206
    lines = []  # type: _coconut.typing.Sequence[str]  # line 206
    if filename is not None:  # line 207
        with open(encode(filename), "rb") as fd:  # line 208
            content = fd.read()  # line 208
    encoding = (lambda _coconut_none_coalesce_item: sys.getdefaultencoding() if _coconut_none_coalesce_item is None else _coconut_none_coalesce_item)(detectEncoding(content))  # line 209
    eol = eoldet(content)  # line 210
    if filename is not None:  # line 211
        with codecs.open(encode(filename), encoding=encoding) as fd2:  # line 212
            lines = safeSplit(fd2.read(), ((b"\n" if eol is None else eol)).decode(encoding))  # line 212
    elif content is not None:  # line 213
        lines = safeSplit(content.decode(encoding), ((b"\n" if eol is None else eol)).decode(encoding))  # line 214
    else:  # line 215
        return (sys.getdefaultencoding(), b"\n", [])  # line 215
    return (encoding, eol, lines)  # line 216

@_coconut_tco  # line 218
def dataCopy(_tipe: 'Type[DataType]', _old: 'DataType', *_args, **_kwargs) -> 'DataType':  # line 218
    r = _old._asdict()  # line 218
    r.update(**_kwargs)  # line 218
    return _coconut_tail_call(makedata, _tipe, *(list(_args) + [r[field] for field in _old._fields]))  # line 218

def user_block_input(output: 'List[str]'):  # line 220
    sep = input("Enter end-of-text marker (default: <empty line>: ")  # type: str  # line 221
    line = sep  # type: str  # line 221
    while True:  # line 222
        line = input("> ")  # line 223
        if line == sep:  # line 224
            break  # line 224
        output.append(line)  # line 225

def merge(file: '_coconut.typing.Optional[bytes]'=None, into: '_coconut.typing.Optional[bytes]'=None, filename: '_coconut.typing.Optional[str]'=None, intoname: '_coconut.typing.Optional[str]'=None, mergeOperation: 'MergeOperation'=MergeOperation.BOTH, charMergeOperation: 'MergeOperation'=MergeOperation.BOTH, diffOnly: 'bool'=False, eol: 'bool'=False) -> 'Tuple[Union[bytes, List[MergeBlock]], _coconut.typing.Optional[bytes]]':  # line 227
    ''' Merges other binary text contents 'file' (or reads file 'filename') into current text contents 'into' (or reads file 'intoname'), returning merged result.
      For update, the other version is assumed to be the "new/added" one, while for diff, the current changes are the ones "added".
      However, change direction markers are insert ("+") for elements only in into, and remove ("-") for elements only in other file (just like the diff marks +/-)
      diffOnly returns detected change blocks only, no text merging
      eol flag will use the other file's EOL marks
      in case of replace block and INSERT strategy, the change will be added **behind** the original
  '''  # line 241
    encoding = None  # type: str  # line 242
    othr = None  # type: _coconut.typing.Sequence[str]  # line 242
    othreol = None  # type: _coconut.typing.Optional[bytes]  # line 242
    curr = None  # type: _coconut.typing.Sequence[str]  # line 242
    curreol = None  # type: _coconut.typing.Optional[bytes]  # line 242
    try:  # load files line-wise and normalize line endings (keep the one of the current file) TODO document  # line 243
        encoding, othreol, othr = detectAndLoad(filename=filename, content=file)  # line 244
        encoding, curreol, curr = detectAndLoad(filename=intoname, content=into)  # line 245
    except Exception as E:  # line 246
        Exit("Cannot merge '%s' into '%s': %r" % (filename, intoname, E))  # line 246
    if None not in [othreol, curreol] and othreol != curreol:  # line 247
        warn("Differing EOL-styles detected during merge. Using current file's style for merged output")  # line 247
    output = list(difflib.Differ().compare(othr, curr))  # type: List[str]  # from generator expression  # line 248
    blocks = []  # type: List[MergeBlock]  # merged result in blocks  # line 249
    tmp = []  # type: List[str]  # block lines  # line 250
    last = " "  # type: str  # line 251
    no = None  # type: int  # line 251
    line = None  # type: str  # line 251
    for no, line in enumerate(output + ["X"]):  # EOF marker (difflib's output will never be "X" alone)  # line 252
        if line[0] == last:  # continue filling current block, no matter what type of block it is  # line 253
            tmp.append(line[2:])  # continue filling current block, no matter what type of block it is  # line 253
            continue  # continue filling current block, no matter what type of block it is  # line 253
        if line == "X" and len(tmp) == 0:  # break if nothing left to do, otherwise perform operation for stored block  # line 254
            break  # break if nothing left to do, otherwise perform operation for stored block  # line 254
        if last == " ":  # block is same in both files  # line 255
            if len(tmp) > 0:  # avoid adding empty keep block  # line 256
                blocks.append(MergeBlock(MergeBlockType.KEEP, [line for line in tmp], line=no - len(tmp)))  # avoid adding empty keep block  # line 256
        elif last == "-":  # may be a pure deletion or part of a replacement (with next block being "+")  # line 257
            blocks.append(MergeBlock(MergeBlockType.REMOVE, [line for line in tmp], line=no - len(tmp)))  # line 258
            if len(blocks) >= 2 and blocks[-2].tipe == MergeBlockType.INSERT:  # line 259
                blocks[-2] = MergeBlock(MergeBlockType.REPLACE, blocks[-1].lines, line=no - len(tmp) - 1, replaces=blocks[-2])  # remember replaced stuff with reference to other merge block TODO why -1 necessary?  # line 260
                blocks.pop()  # line 261
        elif last == "+":  # may be insertion or replacement (with previous - block)  # line 262
            blocks.append(MergeBlock(MergeBlockType.INSERT, [line for line in tmp], line=no - len(tmp)))  # first, assume simple insertion, then check for replacement  # line 263
            if len(blocks) >= 2 and blocks[-2].tipe == MergeBlockType.REMOVE:  #  and len(blocks[-1].lines) == len(blocks[-2].lines):  # requires previous block and same number of lines TODO allow multiple intra-line merge for same-length blocks  # line 264
                blocks[-2] = MergeBlock(MergeBlockType.REPLACE, blocks[-2].lines, line=no - len(tmp) - 1, replaces=blocks[-1])  # remember replaced stuff with reference to other merge block TODO why -1 necessary?  # line 265
                blocks.pop()  # remove TOS due to merging two blocks into replace or modify  # line 266
        elif last == "?":  # marker for intra-line change comment -> add to block info  # line 267
            ilm = getIntraLineMarkers(tmp[0])  # type: Range  # TODO still true? "? " line includes a trailing \n for some reason  # line 268
            blocks[-1] = dataCopy(MergeBlock, blocks[-1], changes=ilm)  # line 269
        last = line[0]  # line 270
        tmp[:] = [line[2:]]  # only keep current line for next block  # line 271
# TODO add code to detect block moves here
    nl = othreol if eol else ((othreol if curreol is None else curreol))  # type: bytes  # no default newline, to mark "no newline"  # line 273
    debug("Diff blocks: " + repr(blocks))  # line 274
    if diffOnly:  # line 275
        return (blocks, nl)  # line 275

# now perform merge operations depending on detected blocks
    output[:] = []  # clean list of strings  # line 278
    for block in blocks:  # line 279
        if block.tipe == MergeBlockType.KEEP:  # line 280
            output.extend(block.lines)  # line 281
        elif block.tipe == MergeBlockType.INSERT and not (mergeOperation.value & MergeOperation.REMOVE.value) or block.tipe == MergeBlockType.REMOVE and (mergeOperation.value & MergeOperation.INSERT.value):  # line 282
            output.extend(block.lines)  # line 284
        elif block.tipe == MergeBlockType.REPLACE:  # complete block replacement  # line 285
            if len(block.lines) == len(block.replaces.lines) == 1:  # one-liner  # line 286
                output.append(lineMerge(block.lines[0], block.replaces.lines[0], mergeOperation=charMergeOperation))  # line 287
            elif mergeOperation == MergeOperation.ASK:  # more than one line: needs user input  # line 288
                printo(ajoin("- ", block.replaces.lines, nl="\n"))  # TODO check +/- in update mode, could be swapped  # line 289
                printo(ajoin("+ ", block.lines, nl="\n"))  # line 290
                while True:  # line 291
                    op = input(" Line replacement: *M[I]ne (+), [T]heirs (-), [B]oth, [U]ser input: ").strip().lower()  # type: str  # line 292
                    if op in "tb":  # line 293
                        output.extend(block.lines)  # line 293
                        break  # line 293
                    if op in "ib":  # line 294
                        output.extend(block.replaces.lines)  # line 294
                        break  # line 294
                    if op == "m":  # line 295
                        user_block_input(output)  # line 295
                        break  # line 295
            else:  # more than one line and not ask  # line 296
                if mergeOperation == MergeOperation.REMOVE:  # line 297
                    pass  # line 297
                elif mergeOperation == MergeOperation.BOTH:  # line 298
                    output.extend(block.lines)  # line 298
                elif mergeOperation == MergeOperation.INSERT:  # TODO optionally allow insertion BEFORE or AFTER original (order of these both lines)  # line 299
                    output.extend(list(block.replaces.lines) + list(block.lines))  # TODO optionally allow insertion BEFORE or AFTER original (order of these both lines)  # line 299
#  debug("Merge output: " + "; ".join(output))
    return (((b"\n" if nl is None else nl)).join([line.encode(encoding) for line in output]), nl)  # returning bytes  # line 301
# TODO handle check for more/less lines in found -/+ blocks to find common section and splitting prefix/suffix out

@_coconut_tco  # line 304
def lineMerge(othr: 'str', into: 'str', mergeOperation: 'MergeOperation'=MergeOperation.BOTH, diffOnly: 'bool'=False) -> 'Union[str, List[MergeBlock]]':  # line 304
    ''' Merges string 'othr' into current string 'into'.
      change direction mark is insert for elements only in into, and remove for elements only in file (according to diff marks +/-)
  '''  # line 307
    out = list(difflib.Differ().compare(othr, into))  # type: List[str]  # line 308
    blocks = []  # type: List[MergeBlock]  # line 309
    for i, line in enumerate(out):  # line 310
        if line[0] == "+":  # line 311
            if i + 1 < len(out) and out[i + 1][0] == "+":  # block will continue  # line 312
                if i > 0 and blocks[-1].tipe == MergeBlockType.INSERT:  # middle of + block  # line 313
                    blocks[-1].lines.append(line[2])  # add one more character to the accumulating list  # line 314
                else:  # first + in block  # line 315
                    blocks.append(MergeBlock(MergeBlockType.INSERT, [line[2]], i))  # line 316
            else:  # last line of + block  # line 317
                if i > 0 and blocks[-1].tipe == MergeBlockType.INSERT:  # end of a block  # line 318
                    blocks[-1].lines.append(line[2])  # line 319
                else:  # single line  # line 320
                    blocks.append(MergeBlock(MergeBlockType.INSERT, [line[2]], i))  # line 321
                if i >= 1 and blocks[-2].tipe == MergeBlockType.REMOVE:  # previous - and now last in + block creates a replacement block  # line 322
                    blocks[-2] = MergeBlock(MergeBlockType.REPLACE, blocks[-2].lines, i, replaces=blocks[-1])  # line 323
                    blocks.pop()  # line 323
        elif line[0] == "-":  # line 324
            if i > 0 and blocks[-1].tipe == MergeBlockType.REMOVE:  # part of - block  # line 325
                blocks[-1].lines.append(line[2])  # line 326
            else:  # first in block  # line 327
                blocks.append(MergeBlock(MergeBlockType.REMOVE, [line[2]], i))  # line 328
        elif line[0] == " ":  # line 329
            if i > 0 and blocks[-1].tipe == MergeBlockType.KEEP:  # part of block  # line 330
                blocks[-1].lines.append(line[2])  # line 331
            else:  # first in block  # line 332
                blocks.append(MergeBlock(MergeBlockType.KEEP, [line[2]], i))  # line 333
        else:  # line 334
            raise Exception("Cannot parse diff line %r" % line)  # line 334
    blocks[:] = [dataCopy(MergeBlock, block, lines=["".join(block.lines)], replaces=dataCopy(MergeBlock, block.replaces, lines=["".join(block.replaces.lines)]) if block.replaces else None) for block in blocks]  # line 335
    if diffOnly:  # line 336
        return blocks  # line 336
    out[:] = []  # line 337
    for i, block in enumerate(blocks):  # line 338
        if block.tipe == MergeBlockType.KEEP:  # line 339
            out.extend(block.lines)  # line 339
        elif block.tipe == MergeBlockType.REPLACE:  # line 340
            if mergeOperation == MergeOperation.ASK:  # line 341
                printo(ajoin("- ", othr))  # line 342
                printo("- " + (" " * i) + block.replaces.lines[0])  # line 343
                printo("+ " + (" " * i) + block.lines[0])  # line 344
                printo(ajoin("+ ", into))  # line 345
                while True:  # line 346
                    op = input(" Character replacement: *M[I]ne (+), [T]heirs (-), [B]oth, [U]ser input: ").strip().lower()  # type: str  # line 347
                    if op in "tb":  # line 348
                        out.extend(block.lines)  # line 348
                        break  # line 348
                    if op in "ib":  # line 349
                        out.extend(block.replaces.lines)  # line 349
                        break  # line 349
                    if op == "m":  # line 350
                        user_block_input(out)  # line 350
                        break  # line 350
            else:  # non-interactive  # line 351
                if mergeOperation == MergeOperation.REMOVE:  # line 352
                    pass  # line 352
                elif mergeOperation == MergeOperation.BOTH:  # line 353
                    out.extend(block.lines)  # line 353
                elif mergeOperation == MergeOperation.INSERT:  # line 354
                    out.extend(list(block.replaces.lines) + list(block.lines))  # line 354
        elif block.tipe == MergeBlockType.INSERT and not (mergeOperation.value & MergeOperation.REMOVE.value):  # line 355
            out.extend(block.lines)  # line 355
        elif block.tipe == MergeBlockType.REMOVE and mergeOperation.value & MergeOperation.INSERT.value:  # line 356
            out.extend(block.lines)  # line 356
# TODO ask for insert or remove as well
    return _coconut_tail_call("".join, out)  # line 358

@_coconut_tco  # line 360
def getIntraLineMarkers(line: 'str') -> 'Range':  # line 360
    ''' Return (type, [affected indices]) of "? "-line diff markers ("? " prefix must be removed). difflib never returns mixed markers per line. '''  # line 361
    if "^" in line:  # TODO wrong, needs removal anyway  # line 362
        return _coconut_tail_call(Range, MergeBlockType.REPLACE, [i for i, c in enumerate(line) if c == "^"])  # TODO wrong, needs removal anyway  # line 362
    if "+" in line:  # line 363
        return _coconut_tail_call(Range, MergeBlockType.INSERT, [i for i, c in enumerate(line) if c == "+"])  # line 363
    if "-" in line:  # line 364
        return _coconut_tail_call(Range, MergeBlockType.REMOVE, [i for i, c in enumerate(line) if c == "-"])  # line 364
    return _coconut_tail_call(Range, MergeBlockType.KEEP, [])  # line 365

def findSosVcsBase() -> 'Tuple[_coconut.typing.Optional[str], _coconut.typing.Optional[str], _coconut.typing.Optional[str]]':  # line 367
    ''' Attempts to find sos and legacy VCS base folders. '''  # line 368
    debug("Detecting root folders...")  # line 369
    path = os.getcwd()  # type: str  # start in current folder, check parent until found or stopped  # line 370
    vcs = (None, None)  # type: Tuple[_coconut.typing.Optional[str], _coconut.typing.Optional[str]]  # line 371
    while not os.path.exists(encode(os.path.join(path, metaFolder))):  # line 372
        contents = set(os.listdir(path))  # type: Set[str]  # line 373
        vcss = [executable for folder, executable in vcsFolders.items() if folder in contents]  # type: _coconut.typing.Sequence[str]  # determine VCS type from dot folder  # line 374
        choice = None  # type: _coconut.typing.Optional[str]  # line 375
        if len(vcss) > 1:  # line 376
            choice = SVN if SVN in vcss else vcss[0]  # line 377
            warn("Detected more than one parallel VCS checkouts %r. Falling back to '%s'" % (vcss, choice))  # line 378
        elif len(vcss) > 0:  # line 379
            choice = vcss[0]  # line 379
        if not vcs[0] and choice:  # memorize current repo root  # line 380
            vcs = (path, choice)  # memorize current repo root  # line 380
        new = os.path.dirname(path)  # get parent path  # line 381
        if new == path:  # avoid infinite loop  # line 382
            break  # avoid infinite loop  # line 382
        path = new  # line 383
    if os.path.exists(encode(os.path.join(path, metaFolder))):  # found something  # line 384
        if vcs[0]:  # already detected vcs base and command  # line 385
            return (path, vcs[0], vcs[1])  # already detected vcs base and command  # line 385
        sos = path  # line 386
        while True:  # continue search for VCS base  # line 387
            new = os.path.dirname(path)  # get parent path  # line 388
            if new == path:  # no VCS folder found  # line 389
                return (sos, None, None)  # no VCS folder found  # line 389
            path = new  # line 390
            contents = set(os.listdir(path))  # line 391
            vcss = [executable for folder, executable in vcsFolders.items() if folder in contents]  # determine VCS type  # line 392
            choice = None  # line 393
            if len(vcss) > 1:  # line 394
                choice = SVN if SVN in vcss else vcss[0]  # line 395
                warn("Detected more than one parallel VCS checkouts %r. Falling back to '%s'" % (vcss, choice))  # line 396
            elif len(vcss) > 0:  # line 397
                choice = vcss[0]  # line 397
            if choice:  # line 398
                return (sos, path, choice)  # line 398
    return (None, vcs[0], vcs[1])  # line 399

def tokenizeGlobPattern(pattern: 'str') -> 'List[GlobBlock]':  # line 401
    index = 0  # type: int  # line 402
    out = []  # type: List[GlobBlock]  # literal = True, first index  # line 403
    while index < len(pattern):  # line 404
        if pattern[index:index + 3] in ("[?]", "[*]", "[[]", "[]]"):  # line 405
            out.append(GlobBlock(False, pattern[index:index + 3], index))  # line 405
            continue  # line 405
        if pattern[index] in "*?":  # line 406
            count = 1  # type: int  # line 407
            while index + count < len(pattern) and pattern[index] == "?" and pattern[index + count] == "?":  # line 408
                count += 1  # line 408
            out.append(GlobBlock(False, pattern[index:index + count], index))  # line 409
            index += count  # line 409
            continue  # line 409
        if pattern[index:index + 2] == "[!":  # line 410
            out.append(GlobBlock(False, pattern[index:pattern.index("]", index + 2) + 1], index))  # line 410
            index += len(out[-1][1])  # line 410
            continue  # line 410
        count = 1  # line 411
        while index + count < len(pattern) and pattern[index + count] not in "*?[":  # line 412
            count += 1  # line 412
        out.append(GlobBlock(True, pattern[index:index + count], index))  # line 413
        index += count  # line 413
    return out  # line 414

def tokenizeGlobPatterns(oldPattern: 'str', newPattern: 'str') -> 'Tuple[_coconut.typing.Sequence[GlobBlock], _coconut.typing.Sequence[GlobBlock]]':  # line 416
    ot = tokenizeGlobPattern(oldPattern)  # type: List[GlobBlock]  # line 417
    nt = tokenizeGlobPattern(newPattern)  # type: List[GlobBlock]  # line 418
#  if len(ot) != len(nt): Exit("Source and target patterns can't be translated due to differing number of parsed glob markers and literal strings")
    if len([o for o in ot if not o.isLiteral]) < len([n for n in nt if not n.isLiteral]):  # line 420
        Exit("Source and target file patterns contain differing number of glob markers and can't be translated")  # line 420
    if any((O.content != N.content for O, N in zip([o for o in ot if not o.isLiteral], [n for n in nt if not n.isLiteral]))):  # line 421
        Exit("Source and target file patterns differ in semantics")  # line 421
    return (ot, nt)  # line 422

def convertGlobFiles(filenames: '_coconut.typing.Sequence[str]', oldPattern: '_coconut.typing.Sequence[GlobBlock]', newPattern: '_coconut.typing.Sequence[GlobBlock]') -> '_coconut.typing.Sequence[Tuple[str, str]]':  # line 424
    ''' Converts given filename according to specified file patterns. No support for adjacent glob markers currently. '''  # line 425
    pairs = []  # type: List[Tuple[str, str]]  # line 426
    for filename in filenames:  # line 427
        literals = [l for l in oldPattern if l.isLiteral]  # type: List[GlobBlock]  # source literals  # line 428
        nextliteral = 0  # type: int  # line 429
        parsedOld = []  # type: List[GlobBlock2]  # line 430
        index = 0  # type: int  # line 431
        for part in oldPattern:  # match everything in the old filename  # line 432
            if part.isLiteral:  # line 433
                parsedOld.append(GlobBlock2(True, part.content, part.content))  # line 433
                index += len(part.content)  # line 433
                nextliteral += 1  # line 433
            elif part.content.startswith("?"):  # line 434
                parsedOld.append(GlobBlock2(False, part.content, filename[index:index + len(part.content)]))  # line 434
                index += len(part.content)  # line 434
            elif part.content.startswith("["):  # line 435
                parsedOld.append(GlobBlock2(False, part.content, filename[index]))  # line 435
                index += 1  # line 435
            elif part.content == "*":  # line 436
                if nextliteral >= len(literals):  # line 437
                    parsedOld.append(GlobBlock2(False, part.content, filename[index:]))  # line 437
                    break  # line 437
                nxt = filename.index(literals[nextliteral].content, index)  # type: int  # also matches empty string  # line 438
                parsedOld.append(GlobBlock2(False, part.content, filename[index:nxt]))  # line 439
                index = nxt  # line 439
            else:  # line 440
                Exit("Invalid file pattern specified for move/rename")  # line 440
        globs = [g for g in parsedOld if not g.isLiteral]  # type: List[GlobBlock2]  # line 441
        literals = [l for l in newPattern if l.isLiteral]  # target literals  # line 442
        nextliteral = 0  # line 443
        nextglob = 0  # type: int  # line 443
        outname = []  # type: List[str]  # line 444
        for part in newPattern:  # generate new filename  # line 445
            if part.isLiteral:  # line 446
                outname.append(literals[nextliteral].content)  # line 446
                nextliteral += 1  # line 446
            else:  # line 447
                outname.append(globs[nextglob].matches)  # line 447
                nextglob += 1  # line 447
        pairs.append((filename, "".join(outname)))  # line 448
    return pairs  # line 449

@_coconut_tco  # line 451
def reorderRenameActions(actions: '_coconut.typing.Sequence[Tuple[str, str]]', exitOnConflict: 'bool'=True) -> '_coconut.typing.Sequence[Tuple[str, str]]':  # line 451
    ''' Attempt to put all rename actions into an order that avoids target == source names.
      Note, that it's currently not really possible to specify patterns that make this work (swapping "*" elements with a reference).
      An alternative would be to always have one (or all) files renamed to a temporary name before renaming to target filename.
  '''  # line 455
    if not actions:  # line 456
        return []  # line 456
    sources = None  # type: List[str]  # line 457
    targets = None  # type: List[str]  # line 457
    sources, targets = [list(l) for l in zip(*actions)]  # line 458
    last = len(actions)  # type: int  # line 459
    while last > 1:  # line 460
        clean = True  # type: bool  # line 461
        for i in range(1, last):  # line 462
            try:  # line 463
                index = targets[:i].index(sources[i])  # type: int  # line 464
                sources.insert(index, sources.pop(i))  # bubble up the action right before conflict  # line 465
                targets.insert(index, targets.pop(i))  # line 466
                clean = False  # line 467
            except:  # target not found in sources: good!  # line 468
                continue  # target not found in sources: good!  # line 468
        if clean:  # line 469
            break  # line 469
        last -= 1  # we know that the last entry in the list has the least conflicts, so we can disregard it in the next iteration  # line 470
    if exitOnConflict:  # line 471
        for i in range(1, len(actions)):  # line 472
            if sources[i] in targets[:i]:  # line 473
                Exit("There is no order of renaming actions that avoids copying over not-yet renamed files: '%s' is contained in matching source filenames" % (targets[i]))  # line 473
    return _coconut_tail_call(list, zip(sources, targets))  # line 474

def relativize(root: 'str', filepath: 'str') -> 'Tuple[str, str]':  # line 476
    ''' Determine OS-independent relative folder path, and relative pattern path. '''  # line 477
    relpath = os.path.relpath(os.path.dirname(os.path.abspath(filepath)), root).replace(os.sep, SLASH)  # line 478
    return relpath, os.path.join(relpath, os.path.basename(filepath)).replace(os.sep, SLASH)  # line 479

def parseOnlyOptions(root: 'str', options: '_coconut.typing.Sequence[str]') -> 'Tuple[_coconut.typing.Optional[FrozenSet[str]], _coconut.typing.Optional[FrozenSet[str]]]':  # line 481
    ''' Returns set of --only arguments, and set or --except arguments. '''  # line 482
    cwd = os.getcwd()  # type: str  # line 483
    onlys = []  # type: List[str]  # line 484
    excps = []  # type: List[str]  # line 484
    index = 0  # type: int  # line 484
    while True:  # line 485
        try:  # line 486
            index = 1 + listindex(options, "--only", index)  # line 487
            onlys.append(options[index])  # line 488
        except:  # line 489
            break  # line 489
    index = 0  # line 490
    while True:  # line 491
        try:  # line 492
            index = 1 + listindex(options, "--except", index)  # line 493
            excps.append(options[index])  # line 494
        except:  # line 495
            break  # line 495
    return (frozenset((relativize(root, o)[1] for o in onlys)) if onlys else None, frozenset((relativize(root, e)[1] for e in excps)) if excps else None)  # line 496



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x538a4166

# Compiled with Coconut version 1.3.1-post_dev17 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

sys = _coconut_sys  # line 1

MARKER = r"/SOS/ "  # type: str  # line 3


def usage(appname, version: 'str', short: 'bool'=False):  # line 6
    print("{marker}{appname}{version}".format(marker=MARKER, appname=appname, version="" if not short else " (PyPI: %s)" % version))  # line 7
    if not short:  # line 8
        print("""

Usage: {cmd} <command> [<argument>] [<option1>, ...]        When operating in offline mode, or command is one of "help", "offline", "version"
       {cmd} --sos <sos command and arguments>              When operating in offline mode, forced passthrough to SOS
       {cmd} --vcs <underlying vcs command and arguments>   When operating in offline mode, forced passthrough to traditional VCS
       {cmd} <underlying vcs command and arguments>         When operating in online mode, automatic passthrough to traditional VCS

  Repository handling:
    offline [<name>]                                      Start working offline, creating a branch (named <name>), default name depending on VCS
      --compress                                            Compress versioned files (same as `sos config set compress on && sos offline`)
      --track                                               Setup SVN-style mode: users add/remove tracking patterns per branch
      --picky                                               Setup Git-style mode: users pick files for each operation
      --strict                                              Always compare entire file contents
    online                                                Finish working offline
    dump [<path>/]<name[.sos.zip]>                        Dump entire repository into an archive file

  Working with branches:
    branch [<name>]                                       Create a new branch from current file tree and switch to it
      --last                                                Use last revision, not current file tree, but keep file tree unchanged
      --stay                                                Don't switch to new branch, continue on current one
    destroy [<branch>]                                    Remove (current or specified) branch entirely
    switch [<branch>][/<revision>]                        Continue work on another branch
      --meta                                                Only switch file tracking patterns for current branch, don't update any files
    update [<branch>][/<revision>]                        Integrate work from another branch
      --add       | --rm       | --ask                      Only add new files / only remove vanished files / Ask what to do. Default: add and remove
      --add-lines | --rm-lines | --ask-lines                Only add inserted lines / only remove deleted lines / Ask what to do. Default: add and remove
      --add-chars | --rm-chars | --ask-chars                Only add new characters / only remove vanished characters / Ask what to do. Default: add and remove
      --eol                                                 Use EOL style from the integrated file instead. Default: EOL style of current file

  Working with files:
    commit [<message>]                                    Create a new revision from current state file tree, with an optional commit message
      --tag                                                 Memorizes commit message as a tag that can be used instead of numeric revisions
    diff [<branch>][/<revision>]                          List changes in file tree (or `--from` specified revision) vs. last (or specified) revision
      --to=branch/revision                                  Take "to" revision as target to compare against (instead of current file tree state)
    ls [<folder path>] [--patterns|--tags]                List file tree and mark changes and tracking status

  Defining file patterns:
    add[not] <file pattern>                               Add a tracking pattern to current branch (file pattern). Using addnot adds to tracking blacklist
    rm[not]  <file pattern>                               Remove a tracking pattern. Only useful after "offline --track" or "offline --picky"
    mv[not]  <oldpattern> <newPattern>                    Rename, move, or move and rename tracked files according to tracked file patterns
      --soft                                                Don't move or rename files, only the tracking pattern

  More commands:
    help, --help                                          Show this usage information
    log                                                   List commits of current branch
      --changes                                             Also list file differences
      --diff                                                Also show textual version differences
    status [<branch>][/<revision>]                        Display changed paths vs. last or specified revision on current or specified branch
    status --repo                                         List branches and display repository status
    version                                               Display version and package information

  User configuration:
    config [set/unset/show/add/rm] [<param> [<value>]]    Configure user-global defaults.
                                                          Flags (1/0, on/off, true/false, yes/no):
                                                            strict, track, picky, compress
                                                          Lists (semicolon-separated when set; single values for add/rm):
                                                            texttype, bintype, ignores, ignoreDirs, ignoresWhitelist, ignoreDirsWhitelist
                                                          Supported texts:
                                                            defaultbranch (has a dynamic default value, depending on VCS discovered)
    config show|list flags|lists|texts                    Enumerates all configurable settings for specified type
    config show <key>                                     Displays only single value

  Arguments:
    [<branch>][/<revision>]      Revision string. Branch is optional (defaulting to current branch) and may be a label or number >= 0
                                 Revision is an optional integer and may be negative to reference from the latest commits (-1 is most recent revision), or a tag name

  Common options:
    --force                      Executes potentially harmful operations. SOS will tell you when it needs you to confirm an operation with "--force"
                                   for offline: ignore being already offline, start from scratch (same as 'sos online --force && sos offline')
                                   for online: ignore uncommitted branches, just go online and remove existing offline repository
                                   for commit, switch, update, add: ignore uncommitted changes before executing command
    --strict                     Always perform full content comparison, don't rely only on file size and timestamp
                                   for offline command: memorize strict mode setting in repository
                                   for changes, diff, commit, switch, update, delete: perform operation in strict mode, regardless of repository setting
    --progress                   Display file names during file tree traversal, and show compression advantage, if enabled
    --only   <tracked pattern>   Restrict operation to specified pattern(s). Available for "changes", "commit", "diff", "switch", and "update"
    --except <tracked pattern>   Avoid operation for specified pattern(s). Available for "changes", "commit", "diff", "switch", and "update"
    --{cmd}                      When executing {CMD} not being offline, pass arguments to {CMD} instead (e.g. {cmd} --{cmd} config set key value.)
    --log                        Enable logging details
    --verbose                    Enable verbose output, including show compression ratios""".format(appname=appname, cmd="sos", CMD="SOS"))  # line 87
    sys.exit(0)  # line 88
